# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re


SIGNATURE = '#EXTM3U'
EXT_PREFIX = '#EXT'


class Tags:

    """All possible M3U8 tags."""

    EXT_X_BYTERANGE = 'EXT-X-BYTERANGE'
    EXT_X_TARGETDURATION = 'EXT-X-TARGETDURATION'
    EXT_X_MEDIA_SEQUENCE = 'EXT-X-MEDIA-SEQUENCE'
    EXT_X_KEY = 'EXT-X-KEY'
    EXT_X_PROGRAM_DATE_TIME = 'EXT-X-PROGRAM-DATE-TIME'
    EXT_X_ALLOW_CACHE = 'EXT-X-ALLOW-CACHE'
    EXT_X_PLAYLIST_TYPE = 'EXT-X-PLAYLIST-TYPE'
    EXT_X_ENDLIST = 'EXT-X-ENDLIST'
    EXT_X_MEDIA = 'EXT-X-MEDIA'
    EXT_X_STREAM_INF = 'EXT-X-STREAM-INF'
    EXT_X_DISCONTINUITY = 'EXT-X-DISCONTINUITY'
    EXT_X_I_FRAMES_ONLY = 'EXT-X-I-FRAMES-ONLY'
    EXT_X_I_FRAME_STREAM_INF = 'EXT-X-I-FRAME-STREAM-INF'
    EXT_X_VERSION = 'EXT-X-VERSION'
    EXTINF = 'EXTINF'


class Stream:

    """An M3U8 stream."""

    BANDWIDTH = 'BANDWIDTH'
    PROGRAM_ID = 'PROGRAM-ID'
    CODECS = 'CODECS'
    RESOLUTION = 'RESOLUTION'
    AUDIO = 'AUDIO'
    VIDEO = 'VIDEO'

    def __init__(self):
        self.bandwidth = None
        self.program_id = None
        self.codecs = []
        self.resolution = None
        self.audio = None
        self.video = None
        self.uri = None

    def set_attribute(self, name, value):
        if name == self.BANDWIDTH:
            self.bandwidth = int(value)
        elif name == self.PROGRAM_ID:
            self.program_id = value
        elif name == self.CODECS:
            self.codecs.append(value)
        elif name == self.RESOLUTION:
            self.resolution = value
        elif name == self.AUDIO:
            self.audio = value
        elif name == self.VIDEO:
            self.video = value

    def set_uri(self, uri):
        self.uri = uri


class Key:

    """An M3U8 cryptographic key."""

    METHOD = 'METHOD'
    URI = 'URI'
    IV = 'IV'

    def __init__(self):
        self.method = None
        self.uri = None
        self.iv = None

    def set_attribute(self, name, value):
        if name == self.METHOD:
            self.method = value
        elif name == self.URI:
            self.uri = value
        elif name == self.IV:
            self.iv = value


class Segment:

    """An M3U8 segment."""

    def __init__(self):
        self.key = None
        self.duration = None
        self.title = None
        self.uri = None

    def is_encrypted(self):
        return self.key is not None


class Playlist:

    """An M3U8 playlist."""

    def __init__(self, target_duration, media_sequence, allow_cache,
                 playlist_type, version, streams, segments):
        self.target_duration = target_duration
        self.media_sequence = media_sequence
        self.allow_cache = allow_cache
        self.playlist_type = playlist_type
        self.version = version
        self.streams = streams
        self.segments = segments


def _validate(lines):
    return lines[0].strip() == SIGNATURE


def _get_line_tagname_attributes(line):
    if ':' not in line:
        return (line[1:], '')
    tagname, attributes = line.split(':', 1)

    # Remove the '#'
    tagname = tagname[1:]

    return tagname, attributes


def _line_is_tag(line):
    return line[0:4] == EXT_PREFIX


def _line_is_relative_uri(line):
    return line[0:4] != 'http'


def parse(data, base_uri):
    streams = []
    segments = []
    current_key = None
    allow_cache = False
    target_duration = 0
    media_sequence = 0
    version = 0
    playlist_type = None
    lines = data.split('\n')

    if not _validate(lines):
        raise RuntimeError('Invalid M3U8 file: "{}"'.format(lines[0]))

    for count in range(1, len(lines)):
        line = lines[count]
        if not _line_is_tag(line):
            continue

        tagname, attributes = _get_line_tagname_attributes(line)

        if tagname == Tags.EXT_X_TARGETDURATION:
            target_duration = int(attributes)
        elif tagname == Tags.EXT_X_MEDIA_SEQUENCE:
            media_sequence = int(attributes)
        elif tagname == Tags.EXT_X_KEY:
            current_key = Key()

            # TODO: do not use split since a URL may contain ','
            attributes = attributes.split(',', 1)
            for attribute in attributes:
                name, value = attribute.split('=', 1)
                name = name.strip()
                value = value.strip('"').strip()
                current_key.set_attribute(name, value)
        elif tagname == Tags.EXT_X_ALLOW_CACHE:
            allow_cache = (attributes.strip() == 'YES')
        elif tagname == Tags.EXT_X_PLAYLIST_TYPE:
            playlist_type = attributes.strip()
        elif tagname == Tags.EXT_X_STREAM_INF:
            # Will match <PROGRAM-ID=1,BANDWIDTH=461000,RESOLUTION=480x270,CODECS="avc1.66.30, mp4a.40.5">
            regex = r'([\w-]+=(?:[a-zA-Z0-9]|"[a-zA-Z0-9,. ]*")+),?'
            attributes = re.findall(regex, attributes)

            stream = Stream()
            for attribute in attributes:
                name, value = attribute.split('=')
                name = name.strip()
                value = value.strip()
                stream.set_attribute(name, value)
            stream.uri = lines[count + 1]
            if _line_is_relative_uri(stream.uri):
                stream.uri = '/'.join([base_uri, stream.uri])
            streams.append(stream)
        elif tagname == Tags.EXT_X_VERSION:
            version = attributes
        elif tagname == Tags.EXTINF:
            duration, title = attributes.split(',')
            segment = Segment()
            segment.key = current_key
            segment.duration = float(duration.strip())
            segment.title = title.strip()
            segment.uri = lines[count + 1]
            if _line_is_relative_uri(segment.uri):
                segment.uri = '/'.join([base_uri, segment.uri])
            segments.append(segment)
        else:
            # Ignore as specified in the RFC
            continue

    return Playlist(target_duration, media_sequence, allow_cache,
                    playlist_type, version, streams, segments)



from toutvqt import config
from toutvqt import utils
from toutvqt import __version__


class QTouTvAboutDialog(utils.QCommonDialog, utils.QtUiLoad):
    _UI_NAME = 'about_dialog'

    def __init__(self):
        super().__init__()

        self._setup_ui()

    def _set_version(self):
        self.version_label.setText('v{}'.format(__version__))

    @staticmethod
    def _create_list(alist):
        return '\n'.join(alist)

    def _set_contributors(self):
        contributors = QTouTvAboutDialog._create_list(config.CONTRIBUTORS)
        self.contributors_edit.setPlainText(contributors)

    def _set_contents(self):
        self._set_version()
        self._set_contributors()

    def _setup_ui(self):
        self._load_ui(QTouTvAboutDialog._UI_NAME)
        self._set_contents()
        self.adjustSize()
        self.setFixedWidth(self.width())
        self.setFixedHeight(self.height())



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import requests
import toutv.exceptions
import toutv.mapper
import toutv.config
import toutv.bos as bos


class Transport:

    def get_emissions(self):
        raise NotImplementedError()

    def get_emission_episodes(self, emission_id):
        raise NotImplementedError()

    def get_page_repertoire(self):
        raise NotImplementedError()

    def search_terms(self, query):
        raise NotImplementedError()


class JsonTransport(Transport):

    def __init__(self, proxies=None, auth=None):
        self._mapper = toutv.mapper.JsonMapper()

        self.set_proxies(proxies)
        self.set_auth(auth)

        self._logger = logging.getLogger(self.__class__.__name__)

    def set_proxies(self, proxies):
        self._proxies = proxies

    def set_auth(self, auth):
        self._auth = auth

    def _do_query_url(self, url, params=None, num_tries=1):
        if num_tries > 1:
            timeout = 5
        else:
            timeout = 10

        for i in range(num_tries):
            try:
                return self._do_one_query_url(url, params, timeout)
            except requests.exceptions.Timeout as e:
                if i < num_tries:
                    self._logger.warning("Timeout with %s; will retry...", url)
                else:
                    raise toutv.exceptions.RequestTimeoutError(url, timeout * num_tries) from e

    def _do_one_query_url(self, url, params=None, timeout=10):
        headers = toutv.config.HEADERS

        r = requests.get(url, params=params, headers=headers, proxies=self._proxies, timeout=timeout)
        if r.status_code != 200:
            code = r.status_code
            raise toutv.exceptions.UnexpectedHttpStatusCodeError(url, code)

        return r

    def _do_query_json_url(self, url, params=None, num_tries=1):
        r = self._do_query_url(url, params, num_tries)
        return r.json()

    def _do_query_json_endpoint(self, endpoint, params=None):
        url = '{}{}'.format(toutv.config.TOUTV_JSON_URL_PREFIX, endpoint)
        json = self._do_query_json_url(url, params, num_tries=5)
        return json['d']

    def get_emissions(self):
        # All emissions, including those only available in Extra
        # We don't have much information about them, except their id, title, and URL, but that is enough to be able to fetch them at least.
        url = '{}/presentation/search'.format(toutv.config.TOUTV_BASE_URL)
        params = {'v': 2, 'd': 'android'}
        results_dto = self._do_query_json_url(url, params)

        def dto_to_bo(dto):
            bo = toutv.bos.Emission()

            bo.Title = dto['DisplayText']
            bo.Id = dto['Id']
            bo.Url = dto['Url']

            return bo

        def filter_program(dto):
            return dto['Key'].startswith('program-')

        programs_dto = filter(filter_program, results_dto)
        emissions = map(dto_to_bo, programs_dto)

        return list(emissions)

    def get_emission_episodes(self, emission, short_version=False):
        if short_version:
            episodes = emission.get_episodes()
            if len(episodes) > 0:
                return episodes

        episodes = []

        url = '{}/presentation/{}'.format(toutv.config.TOUTV_BASE_URL, emission.Url)
        params = {'v': 2, 'excludeLineups': False, 'd': 'android'}
        emission_dto = self._do_query_json_url(url, params)
        seasons = emission_dto['SeasonLineups']

        for season in seasons:
            episodes_dto = season['LineupItems']
            for episode_dto in episodes_dto:
                episode = toutv.bos.Episode()
                episode.Title = episode_dto['Title']
                episode.Description = episode_dto['Description']
                if 'Description' in episode_dto['Details']:
                    episode.Description = episode_dto['Details']['Description']
                episode.PID = episode_dto['IdMedia']
                episode.Id = episode_dto['Key'][6:]
                episode.Url = episode_dto['Url']
                episode.AirDateLongString = episode_dto['Details']['AirDate']
                episode.CategoryId = emission.Id
                episode.SeasonAndEpisode = toutv.client.Client._find_last(r'/.*/(.*)$', episode_dto['Url'])
                episode.set_emission(emission)
                episodes.append(episode)

        return episodes

    def get_page_repertoire(self):
        repertoire_dto = self._do_query_json_endpoint('GetPageRepertoire')

        repertoire = bos.Repertoire()

        # Emissions
        if 'Emissions' in repertoire_dto:
            repertoire.Emissions = {}
            emissionrepertoires_dto = repertoire_dto['Emissions']
            for emissionrepertoire_dto in emissionrepertoires_dto:
                er = self._mapper.dto_to_bo(emissionrepertoire_dto,
                                            bos.EmissionRepertoire)
                repertoire.Emissions[er.Id] = er

        # Genre
        if 'Genres' in repertoire_dto:
            # TODO: implement
            pass

        # Country
        if 'Pays' in repertoire_dto:
            # TODO: implement
            pass

        return repertoire

    def search(self, query):
        searchresults = None
        searchresultdatas = []
        params = {'query': query}

        searchresults_dto = self._do_query_json_endpoint('SearchTerms', params)

        searchresults = self._mapper.dto_to_bo(searchresults_dto,
                                               bos.SearchResults)
        if searchresults.Results is not None:
            for searchresultdata_dto in searchresults.Results:
                sr_bo = self._mapper.dto_to_bo(searchresultdata_dto,
                                               bos.SearchResultData)
                searchresultdatas.append(sr_bo)
        searchresults.Results = searchresultdatas

        return searchresults



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# Copyright (c) 2014, Philippe Proulx <eepp.ca>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn OR Philippe Proulx
# BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

USER_AGENT = 'Mozilla/5.0 (iPhone; U; CPU iPhone OS 4_0 like Mac OS X; en-us) AppleWebKit/532.9 (KHTML, like Gecko) Version/4.0.5 Mobile/8A293 Safari/6531.22.7'
HEADERS = {
    'User-Agent': USER_AGENT
}
TOUTV_PLAYLIST_URL = 'http://api.radio-canada.ca/validationMedia/v1/Validation.html'
TOUTV_AUTH_PLAYLIST_URL = 'https://services.radio-canada.ca/media/validation/v2/'
TOUTV_PLAYLIST_PARAMS = {
    'appCode': 'toutv',
    'deviceType': 'iphone4',
    'connectionType': 'wifi',
    'output': 'json'
}
TOUTV_JSON_URL_PREFIX = 'https://api.tou.tv/v1/toutvapiservice.svc/json/'
TOUTV_BASE_URL = 'https://ici.tou.tv'
EMISSION_THUMB_URL_TMPL = 'http://images.tou.tv/w_400,c_scale,r_5/v1/emissions/16x9/{}.jpg'

TOUTV_AUTH_CLIENT_ID = "d6f8e3b1-1f48-45d7-9e28-a25c4c514c60"
TOUTV_AUTH_SESSION_URL = "https://services.radio-canada.ca/auth/oauth/v2/authorize?response_type=token&client_id=%s&scope=media-drmt+oob+openid+profile+email+id.write+media-validation.read.privileged&state=authCode&redirect_uri=http://ici.tou.tv/profiling/callback" % TOUTV_AUTH_CLIENT_ID
TOUTV_AUTH_TOKEN_URL = "https://services.radio-canada.ca/auth/oauth/v2/authorize"
TOUTV_AUTH_LOGIN_URL = "https://services.radio-canada.ca/auth/oauth/v2/authorize/login"
TOUTV_AUTH_CONSENT_URL = "https://services.radio-canada.ca/auth/oauth/v2/authorize/consent"
TOUTV_AUTH_CLAIMS_URL = "https://services.radio-canada.ca/media/validation/v2/GetClaims?token={}"

TOUTV_AUTH_TOKEN_PATH = ".toutv_token"



import logging
from PyQt4 import Qt, QtCore
from toutvqt.emissions_treemodel import EmissionsTreeModelEmission
from toutvqt.emissions_treemodel import EmissionsTreeModelSeason
from toutvqt.emissions_treemodel import EmissionsTreeModelEpisode
from toutvqt.emissions_treemodel import LoadingItem


class QEmissionsTreeViewStyleDelegate(Qt.QStyledItemDelegate):
    def __init__(self):
        super().__init__()

    def paint(self, painter, option, index):
        if type(index.internalPointer()) is LoadingItem:
            option.font.setItalic(True)
        Qt.QStyledItemDelegate.paint(self, painter, option, index)


class QEmissionsTreeView(Qt.QTreeView):
    emission_selected = QtCore.pyqtSignal(object)
    season_selected = QtCore.pyqtSignal(object, int, list)
    episode_selected = QtCore.pyqtSignal(object)
    none_selected = QtCore.pyqtSignal()

    def __init__(self, model):
        super().__init__()

        self._setup(model)

    def _setup(self, model):
        self.setModel(model)
        self.expanded.connect(model.item_expanded)

        selection_model = Qt.QItemSelectionModel(model)
        self.setSelectionModel(selection_model)

        self.setItemDelegate(QEmissionsTreeViewStyleDelegate())

        selection_model.selectionChanged.connect(self.item_selection_changed)

        model.fetching_start.connect(self._on_fetch_start)
        model.fetching_done.connect(self._on_fetch_done)

    def _on_fetch_start(self):
        self.setCursor(QtCore.Qt.WaitCursor)

    def _on_fetch_done(self):
        self.setCursor(QtCore.Qt.ArrowCursor)

    def set_default_columns_widths(self):
        self.setColumnWidth(0, self.width() - 300)
        self.setColumnWidth(1, 100)

    def item_selection_changed(self, selected, deselected):
        logging.debug('Treeview item selection changed')

        indexes = selected.indexes()

        if not indexes:
            self.none_selected.emit()
            return

        index = indexes[0]
        item = index.internalPointer()
        if type(item) == EmissionsTreeModelEmission:
            self.emission_selected.emit(item.bo)
        elif type(item) == EmissionsTreeModelSeason:
            self.season_selected.emit(item.emission.bo, item.number,
                                      item.episodes)
        elif type(item) == EmissionsTreeModelEpisode:
            self.episode_selected.emit(item.bo)
        else:
            self.none_selected.emit()



# Copyright (c) 2014, Philippe Proulx <eepp.ca>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Philippe Proulx BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import shutil

try:
    from termcolor import colored
    _has_termcolor = True
except ImportError:
    _has_termcolor = False


class ProgressBar:
    def __init__(self, filename, segments_count):
        self._filename = filename
        self._segments_count = segments_count

    @staticmethod
    def _get_terminal_width():
        if hasattr(shutil, 'get_terminal_size'):
            return shutil.get_terminal_size()[0]
        else:
            return 80

    def _get_bar_widget(self, width):
        total_segments = self._total_segments
        inner_width = width - 2
        plain = round(total_segments / self._segments_count * inner_width)
        empty = inner_width - plain
        empty_s = '-' * empty

        if _has_termcolor:
            empty_s = colored(empty_s, attrs=['dark'])

        bar = '[{}{}]'.format('#' * plain, empty_s)

        return bar

    def _get_percent_widget(self, width):
        total_segments = self._total_segments
        percent = int(total_segments / self._segments_count * 100)
        base = '{}%'.format(percent)

        return base.rjust(width)

    def _get_segments_widget(self, width):
        total_segments = self._total_segments
        base = '{}/{}'.format(total_segments, self._segments_count)

        return base.rjust(width)

    def _get_size_widget(self, width):
        total_bytes = self._total_bytes

        if total_bytes < (1 << 10):
            base = '{} B'.format(total_bytes)
        elif total_bytes < (1 << 20):
            base = '{:.1f} kiB'.format(total_bytes / (1 << 10))
        elif total_bytes < (1 << 30):
            base = '{:.1f} MiB'.format(total_bytes / (1 << 20))
        else:
            base = '{:.1f} GiB'.format(total_bytes / (1 << 30))

        return base.rjust(width)

    def _get_filename_widget(self, width):
        filename_len = len(self._filename)

        if filename_len < width:
            s = self._filename.ljust(width)
        else:
            s = '{}...'.format(self._filename[:width - 3])

        if _has_termcolor and self._total_segments != self._segments_count:
            s = colored(s, attrs=['bold'])

        return s

    def get_bar(self, total_segments, total_bytes):
        # Different required widths for widgets
        self._total_segments = total_segments
        self._total_bytes = total_bytes
        term_width = ProgressBar._get_terminal_width()
        percent_width = 5
        size_width = 12
        segments_width = len(str(self._segments_count)) * 2 + 4
        padding = 1
        fixed_width = percent_width + size_width + segments_width + padding
        variable_width = term_width - fixed_width
        filename_width = round(variable_width * 0.6)
        bar_width = variable_width - filename_width

        # Get all widgets
        wpercent = self._get_percent_widget(percent_width)
        wsize = self._get_size_widget(size_width)
        wsegments = self._get_segments_widget(segments_width)
        wfilename = self._get_filename_widget(filename_width)
        wbar = self._get_bar_widget(bar_width)

        # Build line
        line = '{}{}{} {}{}'.format(wfilename, wsize, wsegments, wbar,
                                    wpercent)

        return line



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import datetime
import logging
import os
import re
import requests
import toutv.dl
import toutv.config
import toutv.m3u8


def _clean_description(desc):
    desc = desc.replace('\n', ' ')
    desc = desc.replace('  ', ' ')

    return desc.strip()


class _Bo:

    def set_auth(self, auth):
        self._auth = auth

    def get_auth(self):
        if hasattr(self, '_auth'):
            return self._auth

        self._auth = None

        return self._auth

    def set_proxies(self, proxies):
        self._proxies = proxies

    def get_proxies(self):
        if hasattr(self, '_proxies'):
            return self._proxies

        self._proxies = None

        return self._proxies

    def _do_request(self, url, timeout=None, params=None):
        proxies = self.get_proxies()
        auth = self.get_auth()

        try:
            headers = dict(toutv.config.HEADERS)

            if auth and params:
                url = toutv.config.TOUTV_AUTH_PLAYLIST_URL
                token = auth.get_token()
                params['claims'] = auth.get_claims(token)
                headers['Authorization'] = "Bearer " + token
                headers['Host'] = "services.radio-canada.ca"

            r = requests.get(url, params=params, headers=headers,
                             proxies=proxies, timeout=timeout)
            if r.status_code != 200:
                raise toutv.exceptions.UnexpectedHttpStatusCodeError(url,
                                                                     r.status_code)
        except requests.exceptions.Timeout:
            raise toutv.exceptions.RequestTimeoutError(url, timeout)

        return r


class _ThumbnailProvider:

    def _cache_medium_thumb(self):
        if self.has_medium_thumb_data():
            # No need to download again
            return

        urls = self.get_medium_thumb_urls()

        for url in urls:
            if not url:
                continue

            logging.debug('HTTP-getting "{}"'.format(url))

            try:
                r = self._do_request(url, timeout=2)
            except Exception as e:
                # Ignore any network error
                logging.warning(e)
                continue

            self._medium_thumb_data = r.content
            break

    def get_medium_thumb_data(self):
        self._cache_medium_thumb()

        return self._medium_thumb_data

    def has_medium_thumb_data(self):
        if not hasattr(self, '_medium_thumb_data'):
            self._medium_thumb_data = None

        return (self._medium_thumb_data is not None)

    def get_medium_thumb_urls(self):
        """Returns a list of possible thumbnail urls in order of preference."""
        raise NotImplementedError()


class _AbstractEmission(_Bo):

    def get_id(self):
        return self.Id

    def get_genre(self):
        return self.Genre

    def get_url(self):
        if self.Url is None:
            return None

        return '{}/{}'.format(toutv.config.TOUTV_BASE_URL, self.Url)

    def get_removal_date(self):
        if self.DateRetraitOuEmbargo is None:
            return None

        # Format looks like: /Date(1395547200000-0400)/
        # Sometimes we have weird values: '/Date(-62135578800000-0500)/',
        # we'll return None in that case.
        d = self.DateRetraitOuEmbargo
        m = re.match(r'/Date\((\d+)-\d+\)/', d)
        if m is not None:
            ts = int(m.group(1)) // 1000

            return datetime.datetime.fromtimestamp(ts)

        return None

    def __str__(self):
        return '{} ({})'.format(self.get_title(), self.get_id())


class Emission(_AbstractEmission, _ThumbnailProvider):

    def __init__(self):
        self.CategoryURL = None
        self.ClassCategory = None
        self.ContainsAds = None
        self.Country = None
        self.DateRetraitOuEmbargo = None
        self.Description = None
        self.DescriptionOffline = None
        self.DescriptionUnavailable = None
        self.DescriptionUnavailableText = None
        self.DescriptionUpcoming = None
        self.DescriptionUpcomingText = None
        self.EstContenuJeunesse = None
        self.EstExclusiviteRogers = None
        self.GeoTargeting = None
        self.Genre = None
        self.Id = None
        self.ImageBackground = None
        self.ImagePromoLargeI = None
        self.ImagePromoLargeJ = None
        self.ImagePromoNormalK = None
        self.Network = None
        self.Network2 = None
        self.Network3 = None
        self.ParentId = None
        self.Partner = None
        self.PlaylistExist = None
        self.PromoDescription = None
        self.PromoTitle = None
        self.RelatedURL1 = None
        self.RelatedURL2 = None
        self.RelatedURL3 = None
        self.RelatedURL4 = None
        self.RelatedURL5 = None
        self.RelatedURLImage1 = None
        self.RelatedURLImage2 = None
        self.RelatedURLImage3 = None
        self.RelatedURLImage4 = None
        self.RelatedURLImage5 = None
        self.RelatedURLText1 = None
        self.RelatedURLText2 = None
        self.RelatedURLText3 = None
        self.RelatedURLText4 = None
        self.RelatedURLText5 = None
        self.SeasonNumber = None
        self.Show = None
        self.ShowSearch = None
        self.SortField = None
        self.SortOrder = None
        self.SubCategoryType = None
        self.Title = None
        self.TitleIndex = None
        self.Url = None
        self.Year = None
        self._episodes = {}

    def add_episode(self, episode):
        self._episodes[episode.Id] = episode

    def get_episodes(self):
        return self._episodes.values()

    def get_title(self):
        return self.Title

    def get_year(self):
        return self.Year

    def get_country(self):
        return self.Country

    def get_description(self):
        if self.Description is None:
            return None
        return _clean_description(self.Description)

    def get_network(self):
        if self.Network == '(not specified)':
            return None

        if self.Network is None:
            # We observed CBFT (SRC) is the default network when not specified
            return 'CBFT'

        return self.Network

    def get_tags(self):
        tags = []
        if self.EstExclusiviteRogers:
            tags.append('rogers')
        if self.EstContenuJeunesse:
            tags.append('youth')

        return tags

    def get_medium_thumb_urls(self):
        name = self.Url.replace('-', '')
        url = toutv.config.EMISSION_THUMB_URL_TMPL.format(name)

        return [url, self.ImagePromoNormalK]


class Genre(_Bo):

    def __init__(self):
        self.CategoryURL = None
        self.ClassCategory = None
        self.Description = None
        self.Id = None
        self.ImageBackground = None
        self.ParentId = None
        self.Title = None
        self.Url = None

    def get_id(self):
        return self.Id

    def get_title(self):
        return self.Title

    def __str__(self):
        return '{} ({})'.format(self.get_title(), self.get_id())


class Episode(_Bo, _ThumbnailProvider):

    class Quality:

        def __init__(self, bitrate, xres, yres):
            self._bitrate = bitrate
            self._xres = xres
            self._yres = yres

        @property
        def bitrate(self):
            return self._bitrate

        @property
        def xres(self):
            return self._xres

        @property
        def yres(self):
            return self._yres

        def __hash__(self):
            return hash(self._bitrate) + hash(self._xres) + hash(self._yres)

        def __eq__(self, other):
            return self.bitrate == other.bitrate and self.xres == other.xres \
                and self.yres == other.yres

        def __repr__(self):
            s = 'Quality(res={xres}x{yres}, bitrate={bitrate})'
            return s.format(xres=self.xres,
                            yres=self.yres,
                            bitrate=self.bitrate)

    def __init__(self):
        self.AdPattern = None
        self.AirDateFormated = None
        self.AirDateLongString = None
        self.Captions = None
        self.CategoryId = None
        self.ChapterStartTimes = None
        self.ClipType = None
        self.Copyright = None
        self.Country = None
        self.DateSeasonEpisode = None
        self.Description = None
        self.DescriptionShort = None
        self.EpisodeNumber = None
        self.EstContenuJeunesse = None
        self.Event = None
        self.EventDate = None
        self.FullTitle = None
        self.GenreTitle = None
        self.Id = None
        self.ImageBackground = None
        self.ImagePlayerLargeA = None
        self.ImagePlayerNormalC = None
        self.ImagePromoLargeI = None
        self.ImagePromoLargeJ = None
        self.ImagePromoNormalK = None
        self.ImageThumbMicroG = None
        self.ImageThumbMoyenL = None
        self.ImageThumbNormalF = None
        self.IsMostRecent = None
        self.IsUniqueEpisode = None
        self.Keywords = None
        self.LanguageCloseCaption = None
        self.Length = None
        self.LengthSpan = None
        self.LengthStats = None
        self.LengthString = None
        self.LiveOnDemand = None
        self.MigrationDate = None
        self.Musique = None
        self.Network = None
        self.Network2 = None
        self.Network3 = None
        self.NextEpisodeDate = None
        self.OriginalAirDate = None
        self.PID = None
        self.Partner = None
        self.PeopleAuthor = None
        self.PeopleCharacters = None
        self.PeopleCollaborator = None
        self.PeopleColumnist = None
        self.PeopleComedian = None
        self.PeopleDesigner = None
        self.PeopleDirector = None
        self.PeopleGuest = None
        self.PeopleHost = None
        self.PeopleJournalist = None
        self.PeoplePerformer = None
        self.PeoplePersonCited = None
        self.PeopleSpeaker = None
        self.PeopleWriter = None
        self.PromoDescription = None
        self.PromoTitle = None
        self.Rating = None
        self.RelatedURL1 = None
        self.RelatedURL2 = None
        self.RelatedURL3 = None
        self.RelatedURL4 = None
        self.RelatedURL5 = None
        self.RelatedURLText1 = None
        self.RelatedURLText2 = None
        self.RelatedURLText3 = None
        self.RelatedURLText4 = None
        self.RelatedURLText5 = None
        self.RelatedURLimage1 = None
        self.RelatedURLimage2 = None
        self.RelatedURLimage3 = None
        self.RelatedURLimage4 = None
        self.RelatedURLimage5 = None
        self.SeasonAndEpisode = None
        self.SeasonAndEpisodeLong = None
        self.SeasonNumber = None
        self.Show = None
        self.ShowSearch = None
        self.ShowSeasonSearch = None
        self.StatusMedia = None
        self.Subtitle = None
        self.Team1CountryCode = None
        self.Team2CountryCode = None
        self.Title = None
        self.TitleID = None
        self.TitleSearch = None
        self.Url = None
        self.UrlEmission = None
        self.Year = None
        self.iTunesLinkUrl = None
        self._playlist = None
        self._cookies = None

    def get_title(self):
        return self.Title

    def get_id(self):
        return self.Id

    def get_author(self):
        return self.PeopleAuthor

    def get_director(self):
        return self.PeopleDirector

    def get_year(self):
        return self.Year

    def get_genre_title(self):
        return self.GenreTitle

    def get_url(self):
        if self.Url is None:
            return None

        return '{}/{}'.format(toutv.config.TOUTV_BASE_URL, self.Url).replace('tou.tv//', 'tou.tv/')

    def get_season_number(self):
        return self.SeasonNumber

    def get_episode_number(self):
        return self.EpisodeNumber

    def get_sae(self):
        return self.SeasonAndEpisode

    def get_description(self):
        if self.Description is None:
            return None
        return _clean_description(self.Description)

    def get_emission_id(self):
        return self.CategoryId

    def get_length(self):
        tot_seconds = int(self.Length) // 1000
        minutes = tot_seconds // 60
        seconds = tot_seconds - (60 * minutes)

        return minutes, seconds

    def get_air_date(self):
        if self.AirDateFormated is None:
            return self.AirDateLongString

        dt = datetime.datetime.strptime(self.AirDateFormated, '%Y%m%d')

        return dt.date()

    def set_emission(self, emission):
        self._emission = emission

    def get_emission(self):
        return self._emission

    @staticmethod
    def _get_video_qualities(playlist):
        qualities = []

        for stream in playlist.streams:
            # TOU.TV team doesnt use the "AUDIO" or "VIDEO" M3U8 tags so
            # we must parse the URL to find out about video stream:
            #   index_X_av.m3u8 -> audio-video (av)
            #   index_X_a.m3u8 -> audio (a)
            if not re.search(r'_av\.m3u8', stream.uri):
                continue

            xres = None
            yres = None

            if stream.resolution is not None:
                m = re.match(r'(\d+)x(\d+)', stream.resolution)

                if m:
                    xres = int(m.group(1))
                    yres = int(m.group(2))

            bw = int(stream.bandwidth)
            quality = Episode.Quality(bw, xres, yres)

            qualities.append(quality)

        return qualities

    def _get_playlist_url(self):
        url = toutv.config.TOUTV_PLAYLIST_URL
        params = dict(toutv.config.TOUTV_PLAYLIST_PARAMS)
        params['idMedia'] = self.PID

        num_tries = 3

        for i in range(num_tries):
            r = self._do_request(url, params=params)
            try:
                response_obj = r.json()
                if response_obj['errorCode']:
                    raise RuntimeError(response_obj['message'])
                return response_obj['url']
            except ValueError as e:
                if i + 1 < num_tries:
                    logging.warning("GetPlaylistURL failed. Will retry...")
                else:
                    raise RuntimeError("Error: GetPlaylistURL failed.") from e

    def get_playlist_cookies(self):
        if not self._playlist or not self._cookies:
            url = self._get_playlist_url()
            r = self._do_request(url)

            # parse M3U8 file
            m3u8_file = r.text
            self._playlist = toutv.m3u8.parse(m3u8_file, os.path.dirname(url))
            self._cookies = r.cookies

        return self._playlist, self._cookies

    def get_available_qualities(self):
        # Get playlist
        playlist, cookies = self.get_playlist_cookies()

        # Get video qualities
        qualities = Episode._get_video_qualities(playlist)

        qualities.sort(key=lambda q: q.bitrate)

        return qualities

    def get_medium_thumb_urls(self):
        return [self.ImageThumbMoyenL]

    def __str__(self):
        return '{} ({})'.format(self.get_title(), self.get_id())


class EmissionRepertoire(_AbstractEmission):

    def __init__(self):
        self.AnneeProduction = None
        self.CategorieDuree = None
        self.DateArrivee = None
        self.DateDepart = None
        self.DateRetraitOuEmbargo = None
        self.DescriptionUnavailableText = None
        self.DescriptionUpcomingText = None
        self.Genre = None
        self.Id = None
        self.ImagePromoNormalK = None
        self.IsGeolocalise = None
        self.NombreEpisodes = None
        self.NombreSaisons = None
        self.ParentId = None
        self.Pays = None
        self.SaisonsDisponibles = None
        self.Titre = None
        self.TitreIndex = None
        self.Url = None

    def get_title(self):
        return self.Titre

    def get_country(self):
        return self.Pays

    def get_year(self):
        return self.AnneeProduction


class SearchResults(_Bo):

    def __init__(self):
        self.ModifiedQuery = None
        self.Results = None

    def get_modified_query(self):
        return self.ModifiedQuery

    def get_results(self):
        return self.Results


class SearchResultData(_Bo):

    def __init__(self):
        self.Emission = None
        self.Episode = None

    def get_emission(self):
        return self.Emission

    def get_episode(self):
        return self.Episode


class Repertoire(_Bo):

    def __init__(self):
        self.Emissions = None
        self.Genres = None
        self.Pays = None

    def set_emissions(self, emissions):
        self.Emissions = emissions

    def get_emissions(self):
        return self.Emissions



import queue
import logging
from PyQt4 import Qt
from PyQt4 import QtCore
from toutv import dl


class _DownloadWork:
    def __init__(self, episode, quality, output_dir, proxies):
        self._episode = episode
        self._quality = quality
        self._output_dir = output_dir
        self._proxies = proxies
        self._cancelled = False

    def get_episode(self):
        return self._episode

    @property
    def quality(self):
        return self._quality

    def get_output_dir(self):
        return self._output_dir

    def get_proxies(self):
        return self._proxies

    def cancel(self):
        self._cancelled = True

    def is_cancelled(self):
        return self._cancelled


class _DownloadWorkProgress:
    def __init__(self, done_segments=0, done_bytes=0, done_segments_bytes=0):
        self._done_segments = done_segments
        self._done_bytes = done_bytes
        self._done_segments_bytes = done_segments_bytes

    def get_done_segments(self):
        return self._done_segments

    def get_done_bytes(self):
        return self._done_bytes

    def get_done_segments_bytes(self):
        return self._done_segments_bytes


class _QDownloadStartEvent(Qt.QEvent):
    """Event sent to download workers to make them initiate a download."""

    def __init__(self, type, work):
        super().__init__(type)

        self._work = work

    def get_work(self):
        return self._work


class _QDownloadWorker(Qt.QObject):
    download_started = QtCore.pyqtSignal(object, object, str, int)
    download_progress = QtCore.pyqtSignal(object, object)
    download_finished = QtCore.pyqtSignal(object)
    download_cancelled = QtCore.pyqtSignal(object)
    download_error = QtCore.pyqtSignal(object, object)

    def __init__(self, download_event_type, i):
        super().__init__()
        self._download_event_type = download_event_type
        self._current_work = None
        self._downloader = None
        self._cancelled = False

    def cancel_current_work(self):
        if self._downloader is not None:
            episode = self._current_work.get_episode()
            bitrate = self._current_work.quality.bitrate
            tmpl = 'Cancelling download of "{}" @ {} bps'
            logging.debug(tmpl.format(episode.get_title(), bitrate))
            self._downloader.cancel()

    def cancel_all_works(self):
        self._cancelled = True
        self.cancel_current_work()

    def do_work(self, work):
        if self._cancelled:
            return

        if work.is_cancelled():
            return

        self._current_work = work

        episode = work.get_episode()
        bitrate = work.quality.bitrate
        output_dir = work.get_output_dir()
        proxies = work.get_proxies()

        downloader = dl.Downloader(episode, bitrate=bitrate,
                                   output_dir=output_dir,
                                   on_dl_start=self._on_dl_start,
                                   on_progress_update=self._on_progress_update,
                                   overwrite=True, proxies=proxies)
        self._downloader = downloader

        tmpl = 'Starting download of "{}" @ {} bps'
        logging.debug(tmpl.format(episode.get_title(), bitrate))
        try:
            downloader.download()
        except dl.CancelledByUserError as e:
            self._downloader = None
            self.download_cancelled.emit(work)
            return
        except Exception as e:
            self._downloader = None
            title = episode.get_title()
            tmpl = 'Cannot download "{}" @ {} bps: {}'
            logging.error(tmpl.format(title, bitrate, e))
            self.download_error.emit(work, e)
            return

        self._downloader = None
        self.download_finished.emit(work)

    def _on_dl_start(self, filename, total_segments):
        progress = _DownloadWorkProgress()
        self.download_started.emit(self._current_work, progress, filename,
                                   total_segments)

    def _on_progress_update(self, done_segments, done_bytes,
                            done_segments_bytes):
        dl_progress = _DownloadWorkProgress(done_segments, done_bytes,
                                            done_segments_bytes)
        self.download_progress.emit(self._current_work, dl_progress)

    def _handle_download_event(self, ev):
        self.do_work(ev.get_work())

    def customEvent(self, ev):
        if ev.type() == self._download_event_type:
            self._handle_download_event(ev)
        else:
            logging.error('Download worker received wrong custom event')


class QDownloadManager(Qt.QObject):
    download_created = QtCore.pyqtSignal(object)
    download_started = QtCore.pyqtSignal(object, object, str, int)
    download_progress = QtCore.pyqtSignal(object, object)
    download_finished = QtCore.pyqtSignal(object)
    download_error = QtCore.pyqtSignal(object, object)
    download_cancelled = QtCore.pyqtSignal(object)

    def __init__(self, nb_threads=5):
        super().__init__()

        self._download_event_type = Qt.QEvent.registerEventType()
        self._setup_threads(nb_threads)

    def exit(self):
        # Cancel all workers
        logging.debug('Cancelling all download workers')
        for worker in self._workers:
            worker.cancel_all_works()

        # Clear works
        logging.debug('Clearing remaining download works')
        while not self._works.empty():
            self._works.get()

        # Join threads
        for thread in self._threads:
            logging.debug('Joining one download thread')
            thread.quit()
            thread.wait()

    def cancel_work(self, work):
        if work not in self._works_workers:
            work.cancel()
            self.download_cancelled.emit(work)
        else:
            worker = self._works_workers[work]
            worker.cancel_current_work()

    def _setup_threads(self, nb_threads):
        self._available_workers = queue.Queue()
        self._threads = []
        self._workers = []
        self._works_workers = {}
        self._works = queue.Queue()

        for i in range(nb_threads):
            thread = Qt.QThread()
            worker = _QDownloadWorker(self._download_event_type, i)
            self._threads.append(thread)
            self._workers.append(worker)
            self._available_workers.put(worker)
            worker.moveToThread(thread)
            worker.download_finished.connect(self._on_worker_finished)
            worker.download_error.connect(self._on_worker_error)
            worker.download_cancelled.connect(self._on_worker_finished)

            # Connect worker's signals directly to our signals
            worker.download_cancelled.connect(self.download_cancelled)
            worker.download_error.connect(self.download_error)
            worker.download_finished.connect(self.download_finished)
            worker.download_started.connect(self.download_started)
            worker.download_progress.connect(self.download_progress)

            thread.start()

    def _do_next_work(self):
        try:
            worker = self._available_workers.get_nowait()
        except queue.Empty:
            return

        try:
            work = self._works.get_nowait()
        except queue.Empty:
            self._available_workers.put(worker)
            return

        self._works_workers[work] = worker
        ev = _QDownloadStartEvent(self._download_event_type, work)
        Qt.QCoreApplication.postEvent(worker, ev)

    def download(self, episode, quality, output_dir, proxies):
        work = _DownloadWork(episode, quality, output_dir, proxies)

        self.download_created.emit(work)
        self._works.put(work)
        self._do_next_work()

    def _on_worker_finished(self, work):
        title = work.get_episode().get_title()
        br = work.quality.bitrate
        logging.debug('Download of "{}" @ {} bps ended'.format(title, br))
        worker = self.sender()

        del self._works_workers[work]
        self._available_workers.put(worker)
        self._do_next_work()

    def _on_worker_error(self, work, ex):
        self._on_worker_finished(work)



import os.path
from pkg_resources import resource_filename
from PyQt4 import Qt
from PyQt4 import uic
from toutvqt import config


class QCommonDialog(Qt.QDialog):
    def __init__(self):
        super().__init__()

    def show_move(self, pos):
        self.move(pos)
        self.exec()


class QtUiLoad:
    def _load_ui(self, ui_name):
        ui_rel_path = os.path.join(config.UI_DIR, '{}.ui'.format(ui_name))
        ui_path = resource_filename(__name__, ui_rel_path)
        uic.loadUi(ui_path, baseinstance=self)


def get_qicon(name):
    filename = '{}.png'.format(name)
    rel_path = os.path.join(config.ICONS_DIR, filename)
    path = resource_filename(__name__, rel_path)

    return Qt.QIcon(path)



# Copyright (c) 2014, Philippe Proulx <eepp.ca>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Philippe Proulx BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class NetworkError(RuntimeError):

    def __str__(self):
        return 'Network error'


class RequestTimeoutError(NetworkError):

    def __init__(self, url, timeout):
        self._url = url
        self._timeout = timeout

    @property
    def url(self):
        return self._url

    @property
    def timeout(self):
        return self._timeout

    def __str__(self):
        tmpl = 'Request timeout ({} s for "{}")'
        return tmpl.format(self._timeout, self._url)


class UnexpectedHttpStatusCodeError(NetworkError):

    def __init__(self, url, status_code):
        self._url = url
        self._status_code = status_code

    @property
    def url(self):
        return self._url

    @property
    def status_code(self):
        return self._status_code

    def __str__(self):
        tmpl = 'Unexpected HTTP response code {} for "{}"'
        return tmpl.format(self._status_code, self._url)



from PyQt4.Qt import QDir
from PyQt4.Qt import QSettings
from PyQt4 import QtCore
from PyQt4 import Qt
import logging


class SettingsKeys:
    FILES_DOWNLOAD_DIR = 'files/download_directory'
    NETWORK_HTTP_PROXY = 'network/http_proxy'
    DL_DOWNLOAD_SLOTS = 'downloads/download_slots'
    DL_ALWAYS_MAX_QUALITY = 'downloads/always_max_quality'
    DL_REMOVE_FINISHED = 'downloads/remove_finished'


class QTouTvSettings(Qt.QObject):
    _DEFAULT_DOWNLOAD_DIR = QDir.home().absoluteFilePath('TOU.TV Downloads')
    _settings_types = {
        SettingsKeys.FILES_DOWNLOAD_DIR: str,
        SettingsKeys.NETWORK_HTTP_PROXY: str,
        SettingsKeys.DL_DOWNLOAD_SLOTS: int,
        SettingsKeys.DL_ALWAYS_MAX_QUALITY: bool,
        SettingsKeys.DL_REMOVE_FINISHED: bool,
    }
    setting_item_changed = QtCore.pyqtSignal(str, object)

    def __init__(self):
        super().__init__()
        self._fill_defaults()
        self._settings_dict = {}

    def _fill_defaults(self):
        """Fills defaults with sensible default values."""
        self.defaults = {}
        def_dl_dir = QTouTvSettings._DEFAULT_DOWNLOAD_DIR
        self.defaults[SettingsKeys.FILES_DOWNLOAD_DIR] = def_dl_dir
        self.defaults[SettingsKeys.NETWORK_HTTP_PROXY] = ""
        self.defaults[SettingsKeys.DL_DOWNLOAD_SLOTS] = 5
        self.defaults[SettingsKeys.DL_ALWAYS_MAX_QUALITY] = False
        self.defaults[SettingsKeys.DL_REMOVE_FINISHED] = False

    def write_settings(self):
        logging.debug('Writing settings')

        settings = QSettings()
        settings.clear()

        for k in self._settings_dict:
            if k in self.defaults:
                if self._settings_dict[k] != self.defaults[k]:
                    settings.setValue(k, self._settings_dict[k])
            else:
                msg = 'Setting key {} not found in defaults'.format(k)
                logging.warning(msg)
                settings.setValue(k, self._settings_dict[k])

    def read_settings(self):
        logging.debug('Reading settings')

        settings = QSettings()
        read_settings = self.defaults.copy()
        keys = settings.allKeys()

        for k in keys:
            setting_type = QTouTvSettings._settings_types[k]
            read_settings[k] = settings.value(k, type=setting_type)

        self.apply_settings(read_settings)

    def _apply_settings(self, key, new_value):
        """Apply the new value. Return whether the value changed."""

        # If the key did not exist, it "changed".
        if key not in self._settings_dict:
            self._settings_dict[key] = new_value
            return True

        cur_value = self._settings_dict[key]
        if cur_value == new_value:
            return False

        self._settings_dict[key] = new_value

        return True

    def apply_settings(self, new_settings):
        logging.debug('Applying settings')

        for key in new_settings:
            new_value = new_settings[key]
            if self._apply_settings(key, new_value):
                self.setting_item_changed.emit(key, new_value)

        self.write_settings()

    def get_download_directory(self):
        return self._settings_dict[SettingsKeys.FILES_DOWNLOAD_DIR]

    def get_http_proxy(self):
        return self._settings_dict[SettingsKeys.NETWORK_HTTP_PROXY]

    def get_download_slots(self):
        return int(self._settings_dict[SettingsKeys.DL_DOWNLOAD_SLOTS])

    def get_always_max_quality(self):
        return self._settings_dict[SettingsKeys.DL_ALWAYS_MAX_QUALITY]

    def get_remove_finished(self):
        return self._settings_dict[SettingsKeys.DL_REMOVE_FINISHED]

    def debug_print_settings(self):
        print(self._settings_dict)



from PyQt4 import Qt
from PyQt4 import QtCore
from toutvqt import utils
from enum import Enum


class SymbolicQuality(Enum):
    LOWEST = 0
    AVERAGE = 1
    HIGHEST = 2


class _QQualityButton(Qt.QPushButton):

    def __init__(self, quality):
        super().__init__()

        self._quality = quality

        self._setup()

    def _setup(self):
        self.setText(self._get_text())

    def _get_text(self):
        return ''

    @property
    def quality(self):
        return self._quality


class QBitrateResQualityButton(_QQualityButton):

    def _get_text(self):
        return '{} kbps ({}p)'.format(self.quality.bitrate // 1000,
                                      self.quality.yres)


class QSymbolicQualityButton(_QQualityButton):

    def _get_text(self):
        texts = {
            SymbolicQuality.LOWEST: 'Lowest quality',
            SymbolicQuality.AVERAGE: 'Average quality',
            SymbolicQuality.HIGHEST: 'Highest quality',
        }
        return texts[self.quality]


class QChooseBitrateDialog(utils.QCommonDialog, utils.QtUiLoad):
    _UI_NAME = 'choose_bitrate_dialog'

    bitrate_chosen = QtCore.pyqtSignal(object, list)

    def __init__(self, episodes, qualities, btn_class):
        super().__init__()

        self._episodes = episodes
        self._qualities = qualities
        self._btn_class = btn_class

        self._setup_ui()

    def _setup_ui(self):
        self._load_ui(QChooseBitrateDialog._UI_NAME)
        self._populate_qualities_buttons()
        self.adjustSize()
        self.setFixedHeight(self.height())
        self.setFixedWidth(self.width())

    def _populate_qualities_buttons(self):
        for quality in self._qualities:
            btn = self._btn_class(quality)
            btn.clicked.connect(self._on_quality_btn_clicked)
            btn.adjustSize()
            self.buttons_vbox.addWidget(btn)

    def _on_quality_btn_clicked(self):
        btn = self.sender()
        quality = btn.quality
        self.close()
        self.bitrate_chosen.emit(quality, self._episodes)

    def show_move(self, pos):
        super().show_move(pos)



import unittest

from toutvcli import app


class ToutvCliAppTest(unittest.TestCase):

    def setUp(self):
        self._app = app.App([])

    def _testArgParsing(self, first, second, exp_first, exp_second):
        act_first, act_second = self._app._parse_show_episode_from_args(first, second)
        self.assertEqual(act_first, exp_first)
        self.assertEqual(act_second, exp_second)

    def _testArgParsingRaises(self, first, second):
        with self.assertRaises(app.CliError):
            self._app._parse_show_episode_from_args(first, second)

    def testArgParsing(self):
        self._testArgParsing('http://ici.tou.tv/infoman', None,
                             'infoman', None)
        self._testArgParsing('http://ici.tou.tv/infoman/S17E12?lectureauto=1', None,
                             'infoman', 'S17E12')
        self._testArgParsing('infoman', 'S17E12',
                             'infoman', 'S17E12')

        # Wrong domain
        self._testArgParsingRaises('http://www.perdu.com/', None)

        # Extra argument after URL
        self._testArgParsingRaises('http://ici.tou.tv/infoman', 'S17E12')

        # Wrong URL formats
        self._testArgParsingRaises('http://ici.tou.tv/infoman/S17E12/something', None)
        self._testArgParsingRaises('http://ici.tou.tv/', None)



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import toutv.bos as bos


class Mapper:

    def create(self, klass):
        return klass()


class JsonMapper(Mapper):

    def dto_to_bo(self, dto, klass):
        bo = self.create(klass)
        bo_vars = vars(bo)

        for key in bo_vars.keys():
            if key.startswith("_"):
                continue

            value = dto[key]

            if isinstance(value, dict):
                if '__type' not in value:
                    raise RuntimeError('Cannot find "__type" in value')
                typ = value['__type']

                if typ in ['GenreDTO:#RC.Svc.Web.TouTV',
                           'GenreDTO:RC.Svc.Web.TouTV']:
                    value = self.dto_to_bo(value, bos.Genre)
                elif typ in ['EmissionDTO:#RC.Svc.Web.TouTV',
                             'EmissionDTO:RC.Svc.Web.TouTV']:
                    value = self.dto_to_bo(value, bos.Emission)
                elif typ in ['EpisodeDTO:#RC.Svc.Web.TouTV',
                             'EpisodeDTO:RC.Svc.Web.TouTV']:
                    value = self.dto_to_bo(value, bos.Episode)
            setattr(bo, key, value)

        return bo



import glob
import os
import unittest

from toutvcli import app


run_these_tests = 'PYTOUTV_RUN_REAL_TESTS' in os.environ


@unittest.skipIf(not run_these_tests,
                 "skippiing tests against real service")
class ToutvCliAppRealTest(unittest.TestCase):
    """Tou.TV command line tool tests against the real service.

    These tests are testing the Tou.TV CLI tool against the real Tou.TV
    service.  They might not work everywhere, since a lot of content is
    restricted geographically.  Also, because the content might change over
    time, so the tests may become invalid.  For these reasons, they are skipped
    by default.  If you want to run them, define the PYTOUTV_RUN_REAL_TESTS
    environment variable while running the tests.  For example:

      $ PYTOUTV_RUN_REAL_TESTS=1 pytest test_app_real.py
    """

    def _testInfo(self, args):
        args = ['--verbose', 'info'] + args

        self.assertEqual(app.App(args).run(), 0)

    def testInfo(self):
        self._testInfo(['Le show caché 2'])
        self._testInfo(['Infoman'])
        self._testInfo(['http://ici.tou.tv/infoman'])
        self._testInfo(['Infoman', 'S17E12'])
        self._testInfo(['http://ici.tou.tv/infoman/S17E12?lectureauto=1'])

    def _testList(self, args):
        args = ['--verbose', 'list'] + args

        self.assertEqual(app.App(args).run(), 0)

    def testList(self):
        self._testList([])
        self._testList(['Le show caché 2'])
        self._testList(['Infoman'])
        self._testList(['http://ici.tou.tv/infoman'])

    def _testSearch(self, args):
        args = ['--verbose', 'search'] + args

        self.assertEqual(app.App(args).run(), 0)

    def testSearch(self):
        self._testSearch(['Le show caché 2'])
        self._testSearch(['Infoman'])

    def _testFetch(self, expected_filename, args):
        # Delete the file if it exists prior to the test.
        for file in glob.glob(expected_filename):
            os.unlink(file)

        args = ['--verbose', 'fetch', '-qMIN'] + args

        self.assertEqual(app.App(args).run(), 0)

        self.assertTrue(len(glob.glob(expected_filename)) > 0)

        # Be nice and leave no trace.
        for file in glob.glob(expected_filename):
            os.unlink(file)

    def testFetch(self):
        self._testFetch('Coup.de.pouce.*.S2012E01.*.01.qMIN.ts',
                        ['COUP DE POUCE POUR LA PLANÈTE', 'S2012E01'])



import os.path


ORG_NAME = 'pytoutv'
APP_NAME = 'qtoutv'
DAT_DIR = 'dat'
UI_DIR = os.path.join(DAT_DIR, 'ui')
ICONS_DIR = os.path.join(DAT_DIR, 'icons')
CONTRIBUTORS = [
    'Simon Marchi',
    'Philippe Proulx',
    'Benjamin Vanheuverzwijn',
    'Alexandre Vézina',
    'Alexis Dorais-Joncas',
    'Israel Halle',
    'Marc-Étienne M. Leveillé',
    'Simon Carpentier',
]



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
import difflib
import toutv.cache
import toutv.mapper
import toutv.transport
import toutv.config
import toutv.dl


class NoMatchException(Exception):

    def __init__(self, query, candidates=[]):
        self.query = query
        self.candidates = candidates


class ClientError(RuntimeError):

    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return self._msg


class Client:

    def __init__(self, transport=toutv.transport.JsonTransport(),
                 cache=toutv.cache.EmptyCache(), proxies=None, auth=None):
        self._transport = transport
        self._cache = cache

        self.set_proxies(proxies)
        self.set_auth(auth)

    def set_proxies(self, proxies):
        self._proxies = proxies
        self._transport.set_proxies(proxies)

    def set_auth(self, auth):
        self._auth = auth
        self._transport.set_auth(auth)

    def _set_bo_proxies(self, bo):
        bo.set_proxies(self._proxies)

    def _set_bos_proxies(self, bos):
        for bo in bos:
            self._set_bo_proxies(bo)

    def _set_bo_auth(self, bo):
        bo.set_auth(self._auth)

    def _set_bos_auth(self, bos):
        for bo in bos:
            self._set_bo_auth(bo)

    def get_emissions(self):
        emissions = self._cache.get_emissions()
        if emissions is None:
            emissions = self._transport.get_emissions()
            self._cache.set_emissions(emissions)

        self._set_bos_proxies(emissions)
        self._set_bos_auth(emissions)

        return emissions

    def get_emission_episodes(self, emission, short_version=False):
        episodes = None
        if short_version:
            episodes = self._cache.get_emission_episodes(emission)
        if episodes is None:
            episodes = self._transport.get_emission_episodes(emission, short_version)
            if short_version:
                self._cache.set_emission_episodes(emission, episodes)

        self._set_bos_proxies(episodes)
        self._set_bos_auth(episodes)

        return episodes

    def search(self, query):
        search = self._transport.search(query)
        self._set_bo_proxies(search)

        # Add local emissions (to find Extra emissions & episodes)
        emissions = self.get_emissions()
        query_upper = query.upper()
        for emission in emissions:
            if query_upper in emission.get_title().upper():
                try:
                    # Load this emission' episodes, and add those to the search results too
                    episodes = self._transport.get_emission_episodes(emission, True)

                    sr = toutv.bos.SearchResultData()
                    sr.Emission = emission
                    search.Results.append(sr)

                    for episode in episodes:
                        sr = toutv.bos.SearchResultData()
                        sr.Episode = episode
                        search.Results.append(sr)
                except toutv.exceptions.UnexpectedHttpStatusCodeError as e:
                    if e.status_code == 404:
                        # Show returned by search was not found on new API; just skip this result
                        pass
                    else:
                        raise e

        return search

    def get_emission_by_whatever(self, query):
        shows = self.get_emissions()
        query_upper = query.upper()
        # Map candidates to shows, so that when we get a match between the
        # query string and a candidate string, it's easy to go back to the<
        # corresponding show.
        candidates_to_shows = {}

        # Fill candidates
        for show in shows:
            candidates_to_shows[str(show.get_id())] = show
            candidates_to_shows[show.get_title().upper()] = show
            # The Url property looks like '/infoman', so we strip the leading /.
            candidates_to_shows[show.Url.upper()[1:]] = show

        # Get close matches
        close_matches = difflib.get_close_matches(
            query_upper, candidates_to_shows.keys())

        if not close_matches:
            # No match at all
            raise NoMatchException(query)

        # Matches are sorted by how close they look like the query, so if we
        # have an exact match, it should be first in the list.
        first_match = close_matches[0]
        if first_match != query_upper:
            # No exact match
            raise NoMatchException(query, close_matches)

        # There is an exact match! Return the corresponding show.
        return candidates_to_shows[first_match]

    def get_episode_by_name(self, emission, episode_name, short_version=False):
        episodes = self.get_emission_episodes(emission, short_version)
        episode_name_upper = episode_name.upper()
        candidates = []

        for episode in episodes:
            candidates.append(str(episode.get_id()))
            candidates.append(episode.get_title().upper())
            candidates.append(episode.get_sae())

        # Get close matches
        close_matches = difflib.get_close_matches(episode_name_upper,
                                                  candidates)

        # No match at all
        if not close_matches:
            raise NoMatchException(episode_name)

        # No exact match
        if close_matches[0] != episode_name_upper:
            raise NoMatchException(episode_name, close_matches)

        # Got an exact match
        for episode in episodes:
            search_items = [
                str(episode.get_id()),
                episode.get_title().upper(),
                episode.get_sae()
            ]
            if episode_name_upper in search_items:
                return episode

    @staticmethod
    def _find_last(regex, text):
        results = re.findall(regex, text)
        if not results:
            return None

        return results[-1]



import os
import sys
import logging
import platform
from PyQt4 import Qt
from toutvqt.main_window import QTouTvMainWindow
from toutvqt.settings import QTouTvSettings
from toutvqt.settings import SettingsKeys
from toutvqt import config
import toutv.client


class _QTouTvApp(Qt.QApplication):
    def __init__(self, args):
        super().__init__(args)

        self._proxies = None

        self.setOrganizationName(config.ORG_NAME)
        self.setApplicationName(config.APP_NAME)

        self._setup_client()
        self._setup_settings()
        self._setup_ui()
        self._start()

    def get_settings(self):
        return self._settings

    def get_proxies(self):
        return self._proxies

    def stop(self):
        self.main_window.close()

    def _start(self):
        logging.debug('Starting application')
        self.main_window.start()

    def _setup_ui(self):
        self.main_window = QTouTvMainWindow(self, self._client)

        # Connect the signal between main window and the settings
        self.main_window.settings_accepted.connect(
            self._settings.apply_settings)

    def _setup_client(self):
        self._client = toutv.client.Client()

    def _setup_settings(self):
        # Create a default settings
        self._settings = QTouTvSettings()

        # Connect the signal between settings and us
        self._settings.setting_item_changed.connect(self._setting_item_changed)

        # Read the settings from disk
        self._settings.read_settings()

    def _on_setting_http_proxy_changed(self, value):
        proxies = None
        if value is not None:
            value = value.strip()
            if not value:
                proxies = None
            else:
                proxies = {
                    'http': value,
                    'https': value
                }

        self._proxies = proxies
        self._client.set_proxies(proxies)

    def _on_setting_dl_dir_changed(self, value):
        # Create output directory if it doesn't exist
        if not os.path.exists(value):
            logging.debug('Directory "{}" does not exist'.format(value))
            try:
                os.makedirs(value)
            except Exception:
                # Ignore; should fail later
                logging.warning('Cannot create directory "{}"'.format(value))
                pass

    def _setting_item_changed(self, key, value):
        logging.debug('Setting "{}" changed to "{}"'.format(key, value))
        if key == SettingsKeys.NETWORK_HTTP_PROXY:
            self._on_setting_http_proxy_changed(value)
        elif key == SettingsKeys.FILES_DOWNLOAD_DIR:
            self._on_setting_dl_dir_changed(value)


def _register_sigint(app):
    if platform.system() == 'Linux':
        def handler(signal, frame):
            app.stop()

        import signal
        signal.signal(signal.SIGINT, handler)


def _configure_logging():
    fmt = '[thread %(threadName)10s] %(levelname)7s %(asctime)s ' \
          '%(message)s(%(funcName)s@%(filename)s:%(lineno)d)'
    logging.basicConfig(level=logging.WARNING, format=fmt)


def run():
    _configure_logging()
    app = _QTouTvApp(sys.argv)
    _register_sigint(app)

    return app.exec_()


if __name__ == '__main__':
    run()



import os.path
import logging
from PyQt4 import Qt
from PyQt4 import QtCore
from PyQt4 import QtGui
from toutvqt.download_manager import QDownloadManager
from toutvqt.downloads_tablemodel import QDownloadsTableModel
from toutvqt.downloads_tableview import QDownloadsTableView
from toutvqt.emissions_treeview import QEmissionsTreeView
from toutvqt.emissions_treemodel import EmissionsTreeModel
from toutvqt.about_dialog import QTouTvAboutDialog
from toutvqt.preferences_dialog import QTouTvPreferencesDialog
from toutvqt.choose_bitrate_dialog import QChooseBitrateDialog, SymbolicQuality,\
    QSymbolicQualityButton
from toutvqt.choose_bitrate_dialog import QBitrateResQualityButton
from toutvqt.infos_frame import QInfosFrame
from toutvqt import utils
from toutv import exceptions


class QTouTvMainWindow(Qt.QMainWindow, utils.QtUiLoad):
    _UI_NAME = 'main_window'
    settings_accepted = QtCore.pyqtSignal(dict)

    def __init__(self, app, client):
        super().__init__()

        self._app = app
        self._client = client

        self._setup_ui()

    def _add_treeview(self):
        model = EmissionsTreeModel(self._client)
        model.fetching_start.connect(self._on_treeview_fetch_start)
        model.fetching_done.connect(self._on_treeview_fetch_done)
        self._treeview_model = model

        treeview = QEmissionsTreeView(model)
        self.emissions_treeview = treeview
        self.emissions_tab.layout().addWidget(treeview)

    def _add_tableview(self):
        settings = self._app.get_settings()
        nb_threads = settings.get_download_slots()
        self._download_manager = QDownloadManager(nb_threads=nb_threads)

        model = QDownloadsTableModel(self._download_manager)
        model.download_finished.connect(self._on_download_finished)
        model.download_cancelled.connect(self._on_download_finished)
        self._downloads_tableview_model = model

        tableview = QDownloadsTableView(model)
        self.downloads_tableview = tableview
        self.downloads_tab.layout().addWidget(tableview)

    def _add_infos(self):
        self.infos_frame = QInfosFrame(self._client)
        self.infos_frame.select_download.connect(self._on_select_download)
        self.emissions_tab.layout().addWidget(self.infos_frame)
        treeview = self.emissions_treeview
        treeview.emission_selected.connect(self.infos_frame.show_emission)
        treeview.season_selected.connect(self.infos_frame.show_season)
        treeview.episode_selected.connect(self.infos_frame.show_episode)
        treeview.none_selected.connect(self.infos_frame.show_infos_none)

    def _setup_file_menu(self):
        self.quit_action.triggered.connect(self._app.closeAllWindows)
        self.refresh_emissions_action.triggered.connect(
            self._treeview_model.reset)

    def _setup_edit_menu(self):
        self.preferences_action.triggered.connect(
            self._show_preferences_dialog)

    def _setup_help_menu(self):
        self.about_dialog = QTouTvAboutDialog()
        self.about_action.triggered.connect(self._show_about_dialog)

    def _setup_menus(self):
        self._setup_file_menu()
        self._setup_edit_menu()
        self._setup_help_menu()

    def _setup_action_icon(self, action_name):
        action = getattr(self, action_name)
        icon = utils.get_qicon(action_name)
        action.setIcon(icon)

    def _setup_icons(self):
        self.setWindowIcon(utils.get_qicon('toutv'))
        self._setup_action_icon('quit_action')
        self._setup_action_icon('refresh_emissions_action')
        self._setup_action_icon('preferences_action')
        self._setup_action_icon('about_action')

    def _setup_statusbar(self):
        # Hide status bar until implemented
        self.statusbar.hide()

    def _setup_errormsg(self):
        self._error_msg_dialog = Qt.QErrorMessage(self)

    def _setup_ui(self):
        self._load_ui(QTouTvMainWindow._UI_NAME)
        self._setup_icons()
        self._add_treeview()
        self._add_infos()
        self._add_tableview()
        self._setup_menus()
        self._setup_statusbar()
        self._setup_errormsg()

    def closeEvent(self, close_event):
        logging.debug('Closing main window')
        self._set_wait_cursor()
        self.infos_frame.exit()
        self._downloads_tableview_model.exit()
        self._treeview_model.exit()

    def _setup_ui_post_show(self):
        self.emissions_treeview.set_default_columns_widths()

    def start(self):
        logging.debug('Starting main window')
        self.emissions_treeview.model().init_fetch()
        self.show()
        self._setup_ui_post_show()

    def _show_about_dialog(self):
        pos = self.pos()
        pos.setX(pos.x() + 40)
        pos.setY(pos.y() + 40)
        self.about_dialog.show_move(pos)

    def _show_preferences_dialog(self):
        settings = QTouTvPreferencesDialog(self._app.get_settings())
        settings.settings_accepted.connect(self.settings_accepted)
        pos = self.pos()
        pos.setX(pos.x() + 60)
        pos.setY(pos.y() + 60)
        settings.show_move(pos)

    def _set_wait_cursor(self):
        self.setCursor(QtCore.Qt.WaitCursor)

    def _set_normal_cursor(self):
        self.setCursor(QtCore.Qt.ArrowCursor)

    def _on_download_finished(self, work):
        settings = self._app.get_settings()

        if settings.get_remove_finished():
            eid = work.get_episode().get_id()
            self._downloads_tableview_model.remove_episode_id_item(eid)

    def _on_treeview_fetch_start(self):
        self.refresh_emissions_action.setEnabled(False)

    def _on_treeview_fetch_done(self):
        self.refresh_emissions_action.setEnabled(True)

    def _show_choose_bitrate_dialog(self, episodes, qualities, btn_class):
        pos = QtGui.QCursor().pos()
        dialog = QChooseBitrateDialog(
            episodes, qualities, btn_class)
        dialog.bitrate_chosen.connect(self._on_quality_chosen)
        pos.setX(pos.x() - dialog.width())
        pos.setY(pos.y() - dialog.height())
        dialog.show_move(pos)

    def _on_select_download_single(self, episodes):
        logging.debug('Single episode selected for download')

        assert len(episodes) == 1

        ''' Get available qualities '''
        try:
            self._set_wait_cursor()
            qualities = episodes[0].get_available_qualities()
            btn_class = QBitrateResQualityButton

            settings = self._app.get_settings()
            if settings.get_always_max_quality():
                ''' Assume the qualities are ordered from low to high '''
                self._on_quality_chosen(qualities[-1], episodes)
            else:
                self._show_choose_bitrate_dialog(episodes, qualities,
                                                 btn_class)

        except exceptions.UnexpectedHttpStatusCodeError:
            self._error_msg_dialog.showMessage(
                'Could not download episode playlist. It might not be '
                'available yet.')
            return
        finally:
            self._set_normal_cursor()

    def _on_select_download_multi(self, episodes):
        logging.debug('Multiple episodes selected for download')
        qualities = list(SymbolicQuality)
        btn_class = QSymbolicQualityButton

        settings = self._app.get_settings()
        if settings.get_always_max_quality():
            self._on_quality_chosen(SymbolicQuality.HIGHEST, episodes)
        else:
            self._show_choose_bitrate_dialog(episodes, qualities, btn_class)

    def _on_select_download(self, episodes):
        ''' User clicked on the download button '''
        if len(episodes) == 1:
            self._on_select_download_single(episodes)
        elif len(episodes) > 1:
            self._on_select_download_multi(episodes)

    def _start_download(self, episode, quality, output_dir):
        # Fixme: download_item_exists should take the qu ality as well
        if self._downloads_tableview_model.download_item_exists(
                episode.get_id(), quality):
            tmpl = 'Download of episode "{}" @ {} bps already exists'
            logging.info(tmpl.format(episode.get_title(), quality.bitrate))
            return

        self._download_manager.download(episode, quality, output_dir,
                                        proxies=self._app.get_proxies())

    def start_download_episode_single(self, quality, episode, output_dir):
        self._set_wait_cursor()

        tmpl = 'Queueing download of episode "{}" @ {} bps'
        logging.debug(tmpl.format(episode.get_title(), quality.bitrate))
        self._start_download(episode, quality, output_dir)

        self._set_normal_cursor()

    def start_download_episodes_multi(self, symbolic_quality, episodes,
                                      output_dir):
        self._set_wait_cursor()

        try:
            for episode in episodes:
                qualities = episode.get_available_qualities()
                if symbolic_quality == SymbolicQuality.HIGHEST:
                    quality = qualities[-1]
                elif symbolic_quality == SymbolicQuality.LOWEST:
                    quality = qualities[0]
                elif symbolic_quality == SymbolicQuality.AVERAGE:
                    avg = 0
                    for symbolic_quality in qualities:
                        avg += symbolic_quality.bitrate
                    avg /= len(qualities)
                    quality = min(qualities,
                                  key=lambda q: abs(q.bitrate - avg))

                tmpl = 'Queueing download of episode "{}" @ {} bps'
                logging.debug(tmpl.format(episode.get_title(), quality))
                self._start_download(episode, quality, output_dir)

        except exceptions.UnexpectedHttpStatusCodeError:
            self._error_msg_dialog.showMessage(
                'Could not download episode playlist. It might not be '
                'available yet.')
            return
        finally:
            self._set_normal_cursor()

    def _on_quality_chosen(self, quality, episodes):
        ''' user selected a quality from the popup '''
        logging.debug('Quality chosen')

        settings = self._app.get_settings()
        output_dir = settings.get_download_directory()
        if not os.path.isdir(output_dir):
            msg = 'Output directory "{}" does not exist'.format(output_dir)
            logging.error(msg)
            return

        if len(episodes) == 1:
            self.start_download_episode_single(
                quality, episodes[0], output_dir)
        elif len(episodes) > 1:
            self.start_download_episodes_multi(quality, episodes, output_dir)



from PyQt4 import Qt
from PyQt4 import QtCore
import logging
import re
import toutv.client


class FetchState:
    NOPE = 0
    STARTED = 1
    DONE = 2


class LoadingItem:
    def __init__(self, my_parent):
        self.my_parent = my_parent

    def data(self, index, role):
        column = index.column()
        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                return 'Loading...'
            else:
                return ''

    def rowCount(self):
        # The Loading item does not have any child.
        return 0

    def index(self, row, column, createIndex):
        logging.error('Internal error: index() called on LoadingItem')

        return Qt.QModelIndex()

    def parent(self, child, createIndex):
        if self.my_parent is not None:
            return createIndex(0, 0, self.my_parent)
        else:
            return Qt.QModelIndex()


class EmissionsTreeModelEmission:
    def __init__(self, emission_bo, row_in_parent):
        self.bo = emission_bo
        self.seasons = []
        self.loading_item = LoadingItem(self)
        self.row_in_parent = row_in_parent
        self.fetched = FetchState.NOPE

    def data(self, index, role):
        column = index.column()
        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                return self.bo.get_title()
            elif column == 1:
                network = self.bo.get_network()
                if network is None:
                    network = ''
                return network
            return ''

    def rowCount(self):
        if self.fetched == FetchState.DONE:
            return len(self.seasons)
        else:
            # The "Loading" item
            return 1

    def index(self, row, column, createIndex):
        if self.fetched == FetchState.DONE:
            return createIndex(row, column, self.seasons[row])
        else:
            return createIndex(row, column, self.loading_item)

    def parent(self, child, createIndex):
        # An emission is at root level
        return Qt.QModelIndex()

    def should_fetch(self):
        return self.fetched == FetchState.NOPE

    def set_children(self, c):
        self.seasons = c


class EmissionsTreeModelSeason:
    def __init__(self, number, row_in_parent):
        self.number = number
        self.episodes = []
        self.row_in_parent = row_in_parent

    def data(self, index, role):
        column = index.column()
        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                return 'S{:02}'.format(self.number)
            elif column == 1:
                network = self.emission.bo.get_network()
                if network is None:
                    network = ''

                return network

            return ''

    def rowCount(self):
        return len(self.episodes)

    def index(self, row, column, createIndex):
        return createIndex(row, column, self.episodes[row])

    def should_fetch(self):
        return False

    def parent(self, child, createIndex):
        return createIndex(self.row_in_parent, 0, self.emission)


class EmissionsTreeModelEpisode:
    def __init__(self, bo, row_in_parent):
        self.bo = bo
        self.loading_item = LoadingItem(self)
        self.row_in_parent = row_in_parent

    def data(self, index, role):
        column = index.column()
        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                return self.bo.get_title()
            elif column == 1:
                emission = self.bo.get_emission()
                network = emission.get_network()
                if network is None:
                    network = ''

                return network
            elif column == 2:
                episode_number = self.bo.get_episode_number()
                if episode_number is not None:
                    return episode_number

                return ''

            return '?'

    def rowCount(self):
        # An episode does not have any child
        return 0

    def index(self, row, column, createIndex):
        msg = 'Internal error: index() called on EmissionsTreeModelEpisode'
        logging.error(msg)
        return Qt.QModelIndex()

    def parent(self, child, createIndex):
        return createIndex(self.row_in_parent, 0, self.season)


class EmissionsTreeModel(Qt.QAbstractItemModel):
    _HEADER = [
        'Title',
        'Network',
        'Episode number',
    ]
    fetch_required = QtCore.pyqtSignal(object)
    fetching_start = QtCore.pyqtSignal()
    fetching_done = QtCore.pyqtSignal()

    def __init__(self, client):
        super().__init__()
        self.emissions = []
        self.loading_item = LoadingItem(None)

        # Have we fetched the emissions?
        self.fetched = FetchState.NOPE

        # Setup fetch thread and signal connections
        self.fetch_thread = Qt.QThread()
        self.fetch_thread.start()

        self.fetcher = EmissionsTreeModelFetcher(client)
        self.fetcher.moveToThread(self.fetch_thread)
        self.fetch_required.connect(self.fetcher.new_work_piece)
        self.fetcher.fetch_done.connect(self.fetch_done)
        self.fetcher.fetch_error.connect(self.fetch_error)
        self.modelAboutToBeReset.connect(self._on_about_to_reset)
        self.modelReset.connect(self._on_model_reset)

    def exit(self):
        logging.debug('Joining tree model fetch thread')
        self.fetch_thread.quit()
        self.fetch_thread.wait()

    def index(self, row, column, parent=Qt.QModelIndex()):
        """Returns a QModelIndex to represent a cell of a child of parent."""
        if not parent.isValid():
            # Create an index for a emission
            if self.fetched == FetchState.DONE:
                emission = self.emissions[row]
                return self.createIndex(row, column, emission)
            else:
                return self.createIndex(row, column, self.loading_item)
        else:
            return parent.internalPointer().index(row, column,
                                                  self.createIndex)

    def parent(self, child):
        item = child.internalPointer()
        return item.parent(child, self.createIndex)

    def rowCount(self, parent=Qt.QModelIndex()):
        if not parent.isValid():
            if self.fetched == FetchState.DONE:
                return len(self.emissions)
            else:
                # The "Loading" item
                return 1
        else:
            return parent.internalPointer().rowCount()

    def columnCount(self, parent=Qt.QModelIndex()):
        return len(self._HEADER)

    def fetch_done(self, parent, children_list):
        """A fetch work is complete."""

        # Remove the "Loading"
        self.beginRemoveRows(parent, 0, 0)
        if parent.isValid():
            parent.internalPointer().fetched = FetchState.DONE
        else:
            self.fetched = FetchState.DONE
        self.endRemoveRows()

        # Add the actual children
        self.beginInsertRows(parent, 0, len(children_list) - 1)
        if parent.isValid():
            parent.internalPointer().set_children(children_list)
        else:
            self.emissions = children_list
        self.endInsertRows()

        self.fetching_done.emit()

    def fetch_error(self, parent, ex):
        if type(ex) is toutv.client.ClientError:
            # msg = 'Client error: {}'.format(ex)
            pass
        else:
            logging.error('Error: {}'.format(ex))

        self.fetching_done.emit()

    def init_fetch(self, parent=Qt.QModelIndex()):
        logging.debug('Initializing emissions fetching')

        if parent.isValid():
            parent.internalPointer().fetched = FetchState.STARTED
        else:
            self.fetched = FetchState.STARTED

        parent = Qt.QModelIndex(parent)
        self.fetch_required.emit(parent)
        self.fetching_start.emit()

    def item_expanded(self, parent):
        """Slot called when an item in the tree has been expanded."""
        if parent.internalPointer().should_fetch():
            self.init_fetch(parent)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if not index.isValid():
            return None

        return index.internalPointer().data(index, role)

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.DisplayRole:
            return EmissionsTreeModel._HEADER[section]

    def _on_about_to_reset(self):
        if self.fetched == FetchState.DONE:
            self.emissions = []
            self.fetched = FetchState.NOPE

    def _on_model_reset(self):
        if self.fetched == FetchState.NOPE:
            self.init_fetch()


class EmissionsTreeModelFetcher(Qt.QObject):
    fetch_done = QtCore.pyqtSignal(object, list)
    fetch_error = QtCore.pyqtSignal(object, object)

    def __init__(self, client):
        super().__init__()
        self.client = client

    def new_work_piece(self, parent):
        if not parent.isValid():
            self.fetch_emissions(parent)
        elif type(parent.internalPointer()) == EmissionsTreeModelEmission:
            self.fetch_seasons(parent)
        elif type(parent.internalPointer()) == EmissionsTreeModelSeason:
            self.fetch_episodes(parent)

    def fetch_emissions(self, parent):
        def key_func(ekey):
            # Cheap and easy way to sort latin titles (which is the case here)
            emission_title = emissions[ekey].get_title()
            reps = [
                ('[àáâä]', 'a'),
                ('[ÀÁÂÄ]', 'A'),
                ('[éèêë]', 'e'),
                ('[ÉÈÊË]', 'E'),
                ('[íìîï]', 'i'),
                ('[ÍÌÎÏ]', 'I'),
                ('[óòôö]', 'o'),
                ('[ÓÒÔÖ]', 'O'),
                ('[úùûü]', 'u'),
                ('[ÚÙÛÜ]', 'U'),
                ('ç', 'c'),
                ('Ç', 'C'),
            ]
            for regex, rep in reps:
                emission_title = re.sub(regex, rep, emission_title)

            return emission_title.lower()

        logging.debug('Fetching emissions')

        try:
            emissions = self.client.get_page_repertoire().get_emissions()
        except Exception as e:
            self.fetch_error.emit(parent, e)
            return

        # Sort
        emissions_keys = list(emissions.keys())
        emissions_keys.sort(key=key_func)

        emissions_ret = []
        for i, ekey in enumerate(emissions_keys):
            emission = emissions[ekey]
            new_emission = EmissionsTreeModelEmission(emission, i)
            emissions_ret.append(new_emission)

        self.fetch_done.emit(parent, emissions_ret)

    def fetch_seasons(self, parent):
        emission = parent.internalPointer()
        seasons_list = []
        seasons_dict = {}

        logging.debug('Fetching seasons/episodes')

        try:
            episodes = self.client.get_emission_episodes(emission.bo)
        except Exception as e:
            self.fetch_error.emit(parent, e)
            return

        # Sort
        def key_func(ekey):
            return int(episodes[ekey].get_episode_number())
        episodes_keys = list(episodes.keys())
        episodes_keys.sort(key=key_func)

        for key in episodes_keys:
            ep = episodes[key]
            if ep.get_season_number() not in seasons_dict:
                seasons_dict[ep.get_season_number()] = []
            seasons_dict[ep.get_season_number()].append(ep)

        for i, season_number in enumerate(seasons_dict):
            episodes = seasons_dict[season_number]
            new_season = EmissionsTreeModelSeason(season_number, i)
            new_season.emission = emission
            for (j, ep) in enumerate(episodes):
                new_episode = EmissionsTreeModelEpisode(ep, j)
                new_episode.season = new_season
                new_season.episodes.append(new_episode)
            seasons_list.append(new_season)

        self.fetch_done.emit(parent, seasons_list)



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# Copyright (c) 2014, Philippe Proulx <eepp.ca>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn OR Philippe Proulx
# BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import distutils.version
import locale
import os
import sys
import time
import logging
import textwrap
import platform
import getpass
import re
import toutv.dl
import toutv.client
import toutv.cache
import toutv.config
import toutv.auth
import toutv.exceptions
from toutvcli import __version__
from toutvcli.progressbar import ProgressBar
import traceback
from urllib.parse import urlparse


class CliError(RuntimeError):
    pass


class App:
    QUALITY_MIN = 'MIN'
    QUALITY_AVG = 'AVERAGE'
    QUALITY_MAX = 'MAX'

    FETCH_INFO_FIRST_ARG = 'show-or-url'
    FETCH_INFO_SECOND_ARG = 'episode'

    def __init__(self, args):
        self._argparser = self._build_argparser()
        self._args = args
        self._dl = None
        self._stop = False
        self._logger = logging.getLogger(__name__)
        self._toutv_client = None
        self._verbose = False
        self._quiet = False

    def run(self):
        locale.setlocale(locale.LC_ALL, '')

        # Errors are caught here and a corresponding error code is
        # returned. The codes are:
        #
        #   * 0: all okay
        #   * 1: client error
        #   * 2: download error (cancelled, file exists, no space left, etc.)
        #   * 3: network error (timeout, bad HTTP request, etc.)
        #   * 100: unknown error

        args = self._argparser.parse_args(self._args)

        self._verbose = args.verbose

        if 'quiet' in args:
            if args.quiet:
                self._quiet = True
            elif not sys.__stdin__.isatty():
                print("Non-interactive shell; enabling --quiet")
                self._quiet = True

        if 'no_cache' not in args:
            args.no_cache = False

        no_cache = args.no_cache_global or args.no_cache

        if self._verbose:
            logging.basicConfig(level=logging.DEBUG)

        if args.build_client:
            self._toutv_client = self._build_toutv_client(no_cache)

        try:
            args.func(args)
        except toutv.client.ClientError as e:
            print('Client error: {}'.format(e), file=sys.stderr)
            return 1
        except toutv.dl.CancelledByUserError:
            print('Download cancelled by user', file=sys.stderr)
            return 2
        except toutv.dl.FileExistsError as e:
            tmpl = 'Destination file {} exists (use -f to overwrite).'
            print(tmpl.format(e.path), file=sys.stderr)
            return 2
        except toutv.dl.NoSpaceLeftError:
            print('No space left on device while downloading', file=sys.stderr)
            return 2
        except toutv.dl.DownloadError as e:
            print('Download error: {}'.format(e), file=sys.stderr)

            if self._verbose:
                traceback.print_exc()

            return 2
        except toutv.exceptions.RequestTimeoutError as e:
            tmpl = 'Timeout error ({} s for "{}")'
            print(tmpl.format(e.timeout, e.url), file=sys.stderr)
            return 3
        except toutv.exceptions.UnexpectedHttpStatusCodeError as e:
            tmpl = 'Bad HTTP status code ({}) for "{}"'
            if e.status_code == 401 and os.path.exists(App._build_cache_path(toutv.config.TOUTV_AUTH_TOKEN_PATH)):
                print("Invalid auth token; will be deleted. You will need to run 'toutv login [email_address]' to resolve this issue.", file=sys.stderr)
                App._delete_auth()
                return run()
            print(tmpl.format(e.status_code, e.url), file=sys.stderr)
            return 3
        except toutv.exceptions.NetworkError as e:
            print('Network error: {}'.format(e), file=sys.stderr)
            return 3
        except CliError as e:
            print('Command line error: {}'.format(e), file=sys.stderr)
            return 1
        except toutv.client.NoMatchException as e:
            self._handle_no_match_exception(e)
            return 1
        except Exception as e:
            print('Unknown exception: {}: {}'.format(type(e), e),
                  file=sys.stderr)

            if self._verbose:
                traceback.print_exc()

            return 100

        return 0

    @staticmethod
    def _handle_no_match_exception(e):
        print('Cannot find "{}"'.format(e.query))
        if not e.candidates:
            return
        if len(e.candidates) == 1:
            print('Did you mean "{}"?'.format(e.candidates[0]))
        else:
            print('Did you mean one of the following?\n')
            for candidate in e.candidates:
                print('  * {}'.format(candidate))

    def _build_argparser(self):
        p = argparse.ArgumentParser(description='TOU.TV command line client')
        sp = p.add_subparsers(dest='command', help='Commands help')
        sp.required = True

        # version
        p.add_argument('-n', '--no-cache', action='store_true',
                       dest='no_cache_global', help='Disable cache')
        p.add_argument('-v', '--verbose', action='store_true',
                       help='Verbose output')
        p.add_argument('-V', '--version', action='version',
                       version='%(prog)s v{}'.format(__version__))

        # list command
        usage = ('\n\n'
                 '    1. toutv list [options]\n'
                 '    2. toutv list [options] <show>\n')

        desc = '''
List shows (1) or episodes of a show (2).  In form 2, the show can be specified
using its name, url or id (as found with the form 1 of this command).
'''
        pl = sp.add_parser('list',
                           help='List shows or episodes of a show',
                           usage=usage, description=desc)
        pl.add_argument('show', action='store', nargs='?', type=str,
                        help='Name or url of a show')
        pl.add_argument('-n', '--no-cache', action='store_true',
                        help=argparse.SUPPRESS)
        pl.set_defaults(func=self._command_list)
        pl.set_defaults(build_client=True)

        # info command
        usage = ('\n\n'
                 '    1. toutv info [options] <show-url>\n'
                 '    2. toutv info [options] <episode-url>\n'
                 '    3. toutv info [options] <show>\n'
                 '    4. toutv info [options] <show> <episode>\n')

        desc = '''
Print some information about a show (1 and 3) or an episode (2 and 4). In forms
3 and 4, shows can be referred to using the name or the id as found with the
list command. Episodes can be specified using their name, number or id.
'''

        pi = sp.add_parser('info',
                           help='Get information about a show or an episode',
                           usage=usage, description=desc)
        pi.add_argument(App.FETCH_INFO_FIRST_ARG, action='store', type=str,
                        help='Show or URL, depending on the form used.')
        pi.add_argument(App.FETCH_INFO_SECOND_ARG, action='store', nargs='?', type=str,
                        help='Episode, if necessary.')
        pi.add_argument('-n', '--no-cache', action='store_true',
                        help=argparse.SUPPRESS)
        pi.set_defaults(func=self._command_info)
        pi.set_defaults(build_client=True)

        # search command
        ps = sp.add_parser('search',
                           help='Search TOU.TV shows or episodes')
        ps.add_argument('query', action='store', type=str,
                        help='Search query')
        ps.set_defaults(func=self._command_search)
        ps.set_defaults(build_client=True)

        # fetch command
        usage = ('\n\n'
                 '    1. toutv fetch [options] <episode-url>\n'
                 '    2. toutv fetch [options] <show-url>\n'
                 '    3. toutv fetch [options] <show> <episode>\n'
                 '    4. toutv fetch [options] <show>\n')

        desc = '''
Fetch an episode (1 and 3) or all episodes of a show (2 and 4). In forms 3 and
4, shows can be referred to using the name or the id as found with the list
command. The episode can be specified using its name, number or id.
'''

        pf = sp.add_parser('fetch',
                           help='Fetch one or all episodes of a show',
                           usage=usage, description=desc)
        quality_choices = [
            App.QUALITY_MIN,
            App.QUALITY_AVG,
            App.QUALITY_MAX
        ]
        pf.add_argument(App.FETCH_INFO_FIRST_ARG, action='store', type=str,
                        help='Show or URL, depending on the form used.')
        pf.add_argument(App.FETCH_INFO_SECOND_ARG, action='store', type=str, nargs='?',
                        help='Episode, if necessary.')
        pf.add_argument('-b', '--bitrate', action='store', type=int,
                        help='Video bitrate (default: use default quality)')
        pf.add_argument('-d', '--directory', action='store',
                        default=os.getcwd(),
                        help='Output directory (default: CWD)')
        pf.add_argument('-f', '--force', action='store_true',
                        help='Overwrite existing output file')
        pf.add_argument('-n', '--no-cache', action='store_true',
                        help=argparse.SUPPRESS)
        pf.add_argument('-q', '--quality', action='store',
                        default=App.QUALITY_AVG, choices=quality_choices,
                        help='Video quality (default: {})'.format(App.QUALITY_AVG))
        pf.add_argument('-Q', '--quiet', action='store_true',
                        help='Don\'t show progress while downloading')
        pf.set_defaults(func=self._command_fetch)
        pf.set_defaults(build_client=True)

        # clean command
        pc = sp.add_parser('clean', help='Clean temporary downloaded files')
        pc.add_argument('directory', action='store', nargs='?',
                        default=os.getcwd(),
                        help='Directory to clean (default: CWD)')
        pc.set_defaults(func=self._command_clean)
        pc.set_defaults(build_client=False)

        # login command
        pn = sp.add_parser('login', help='Login to have access to extra content')
        pn.add_argument('username', action='store', type=str,
                        help='Tou.tv user email')
        pn.add_argument('password', action='store', nargs='?', type=str,
                        help='Tou.tv user password')
        pn.set_defaults(func=self._command_login)
        pn.set_defaults(build_client=False)

        return p

    @staticmethod
    def _build_cache_path(cache_name):
        cache_path = cache_name
        if platform.system() in ['Linux', 'Darwin']:
            try:
                cache_dir = os.environ['XDG_CACHE_DIR']
                xdg_cache_path = os.path.join(cache_dir, 'toutv')
                if not os.path.exists(xdg_cache_path):
                    os.makedirs(xdg_cache_path)
                cache_path = os.path.join(xdg_cache_path, cache_name)
            except KeyError:
                home_dir = os.environ['HOME']
                home_cache_path = os.path.join(home_dir, '.cache', 'toutv')
                if not os.path.exists(home_cache_path):
                    os.makedirs(home_cache_path)
                cache_path = os.path.join(home_cache_path, cache_name)
        elif platform.system() == 'Windows':
            if 'APPDATA' in os.environ:
                cache_dir = os.environ['APPDATA']
                appdata_cache_path = os.path.join(cache_dir, 'toutv')
                if not os.path.exists(appdata_cache_path):
                    os.makedirs(appdata_cache_path)
                cache_path = os.path.join(appdata_cache_path, cache_name)

        return cache_path

    @staticmethod
    def _build_cache():
        return toutv.cache.ShelveCache(App._build_cache_path('.toutv_cache'))

    @staticmethod
    def _build_auth():
        auth = None

        try:
            with open(App._build_cache_path(toutv.config.TOUTV_AUTH_TOKEN_PATH), 'r') as token_file:
                auth = toutv.auth.Auth(token_file.read())
        except Exception:
            pass

        return auth

    @staticmethod
    def _delete_auth():
        token_file = App._build_cache_path(toutv.config.TOUTV_AUTH_TOKEN_PATH)
        os.remove(token_file)

    def _build_toutv_client(self, no_cache):
        auth = App._build_auth()

        if no_cache:
            cache = toutv.cache.EmptyCache()
        else:
            try:
                cache = App._build_cache()
            except Exception:
                print('Warning: not using cache (multiple instances of toutv?)',
                      file=sys.stderr)

                if self._verbose:
                    traceback.print_exc()

                cache = toutv.cache.EmptyCache()

        return toutv.client.Client(cache=cache, auth=auth)

    def _parse_show_episode_from_args(self, first, second):
        # Parse the arguments used to specify a show or an episode. If first is
        # an URL, we extract the show name and episode number (if applicable)
        # from it.  If not, we consider that first and second are keywords
        # (name, number or id) describing respectively the show and the
        # episode, so they are returned as-is.  No validation (e.g. check if
        # the show/episode exists) is done on the keywords.
        #
        # Here are examples of allowed forms to refer to an episode:
        #
        #   - "http://ici.tou.tv/district-31/S01E56", None
        #   - "District 31", "S01E56"
        #   - "District 31", "Épisode 56"
        #   - "district-31" "S01E56"
        #   - "2424967431" "2425170307"
        #
        #   or to a show:
        #   - "http://ici.tou.tv/district-31", None
        #   - "district-31", None

        # Try to parse the first argument as an URL, this will tell us if we use the URL form or the keyword form.
        url = urlparse(first)

        if url.scheme in ('http', 'https'):
            # It's an URL, so check that there is no extra argument.
            if second is not None:
                raise CliError('A single argument must be passed when using the URL form.')

            # Validate that the URL is using the right domain.
            if url.netloc != 'ici.tou.tv':
                raise CliError('URL from unrecognized domain {}'.format(url.netloc))

            # Cleanup and split the path part (e.g. path='/district-31/S01E56')
            elements = url.path.strip('/').split('/')

            # Validate the number of path elements.
            if len(elements) not in [1, 2] or len(elements[0]) == 0:
                raise CliError('Couldn\'t recognize URL {}'.format(first))

            show_spec = elements[0]
            episode_spec = elements[1] if len(elements) == 2 else None
        else:
            show_spec = first
            episode_spec = second

        if self._verbose:
            tpl = '_parse_show_episode_from_args returns: emission {} and episode {}'
            print(tpl.format(show_spec, episode_spec))

        return show_spec, episode_spec

    def _get_show_episode_from_args(self, first, second):
        # Get the show and episode objects referred to by first and second.
        # See doc of _parse_show_episode_from_args for the acceptable values.
        show_spec, episode_spec = self._parse_show_episode_from_args(first, second)

        show = self._toutv_client.get_emission_by_whatever(show_spec)

        if episode_spec:
            episode = self._toutv_client.get_episode_by_name(show, episode_spec)
        else:
            episode = None

        return show, episode

    def _command_clean(self, args):
        # make sure we have to clean a directory
        if not os.path.isdir(args.directory):
            raise CliError('"{}" is not an existing directory'.format(args.directory))

        import glob

        tmpdl = glob.glob(os.path.join(args.directory, '.toutv-*.*'))
        tmpcomplete = glob.glob(os.path.join(args.directory, '*.ts.part'))

        for f in tmpdl + tmpcomplete:
            try:
                self._logger.debug('removing file "{}"'.format(f))
                os.remove(f)
            except Exception as e:
                # not the end of the world
                self._logger.warning('could not remove file "{}": {}'.format(f, e))

    @staticmethod
    def _command_login(args):
        auth = toutv.auth.Auth()

        password = args.password
        if password is None:
            password = getpass.getpass()

        try:
            auth.login(args.username, password)
            token = auth.get_token()

            print('Login successful')
            print('Token: {}'.format(token))

            with open(App._build_cache_path(toutv.config.TOUTV_AUTH_TOKEN_PATH), 'w') as token_file:
                token_file.write(token)
        except Exception as e:
            print('Unable to login "{}"'.format(e))

    def _command_list(self, args):
        if args.show:
            # List episodes of a show.
            show, episode = self._get_show_episode_from_args(args.show, None)

            if episode:
                raise CliError("Episode URL given to the list command.")

            self._print_list_episodes(show)
        else:
            # List shows.
            self._print_list_emissions()

    def _command_info(self, args):
        first = getattr(args, App.FETCH_INFO_FIRST_ARG)
        second = getattr(args, App.FETCH_INFO_SECOND_ARG)

        show, episode = self._get_show_episode_from_args(first, second)

        if episode:
            self._print_info_episode(episode)
        else:
            self._print_info_emission(show)

    def _command_fetch(self, args):
        output_dir = args.directory
        bitrate = args.bitrate
        quality = args.quality
        overwrite = args.force

        first = getattr(args, App.FETCH_INFO_FIRST_ARG)
        second = getattr(args, App.FETCH_INFO_SECOND_ARG)

        show, episode = self._get_show_episode_from_args(first, second)

        if episode:
            self._fetch_episode(episode, output_dir=output_dir, quality=quality, bitrate=bitrate, overwrite=overwrite)
        else:
            self._fetch_emission_episodes(show, output_dir=output_dir, quality=quality, bitrate=bitrate,
                                          overwrite=overwrite)

    def _command_search(self, args):
        self._print_search_results(args.query)

    def _print_search_results(self, query):
        searchresult = self._toutv_client.search(query)

        modified_query = searchresult.get_modified_query()
        print('Effective query: {}\n'.format(modified_query))

        if not searchresult.get_results():
            print('No results')
            return

        for result in searchresult.get_results():
            if result.get_emission() is not None:
                emission = result.get_emission()
                print('Emission: {}  [{}]'.format(emission.get_title(),
                                                  emission.get_id()))

                if emission.get_description():
                    print('')
                    description = textwrap.wrap(emission.get_description(), 78)
                    for line in description:
                        print('  {}'.format(line))

            if result.get_episode() is not None:
                episode = result.get_episode()
                print('Episode: {}  [{}]'.format(episode.get_title(),
                                                 episode.get_id()))

                infos_lines = []

                air_date = episode.get_air_date()
                if air_date is not None:
                    line = '  * Air date: {}'.format(air_date)
                    infos_lines.append(line)

                emission_id = episode.get_emission_id()
                if emission_id is not None:
                    line = '  * Emission ID: {}'.format(emission_id)
                    infos_lines.append(line)

                if infos_lines:
                    print('')
                    for line in infos_lines:
                        print(line)

                if episode.get_description():
                    print('')
                    description = textwrap.wrap(episode.get_description(), 78)
                    for line in description:
                        print('  {}'.format(line))

            print('\n')

    def _print_list_emissions(self):
        shows = self._toutv_client.get_emissions()

        def title_sort_func(show):
            return locale.strxfrm(show.get_title())

        for show in sorted(shows, key=title_sort_func):
            title = show.get_title()
            print('{} - {}'.format(show.get_id(), title))

    def _print_list_episodes(self, emission):
        episodes = self._toutv_client.get_emission_episodes(emission, True)

        print('{}:\n'.format(emission.get_title()))
        if len(episodes) == 0:
            print('No available episodes')
            return

        for episode in App._sort_episodes(episodes):
            sae = episode.get_sae()
            title = episode.get_title()
            date = episode.get_air_date()
            print('  * {} - {} - {}'.format(sae, title, date))

    def _print_info_emission(self, emission):
        if emission.get_description() is None:
            url = '{}/presentation{}'.format(toutv.config.TOUTV_BASE_URL, emission.Url)
            emission_dto = self._toutv_client._transport._do_query_json_url(url, {'v': 2, 'excludeLineups': True, 'd': 'android'})
            emission.Description = emission_dto['Details']['Description']
            emission.Country = emission_dto['Details']['Country']

        inner = emission.get_country()
        if inner is None:
            inner = 'Unknown country'
        if emission.get_year() is not None:
            inner = '{}, {}'.format(inner, emission.get_year())
        print('{}  [{}]'.format(emission.get_title(), inner))
        print('ID: {}'.format(emission.get_id()))

        if emission.get_description() is not None:
            print('')
            description = textwrap.wrap(emission.get_description(), 80)
            for line in description:
                print(line)

        infos_lines = []
        if emission.get_network() is not None:
            line = '  * Network: {}'.format(emission.get_network())
            infos_lines.append(line)
        removal_date = emission.get_removal_date()
        if removal_date is not None:
            line = '  * Removal date: {}'.format(removal_date)
            infos_lines.append(line)
        tags = emission.get_tags()
        if tags:
            tags_list = ', '.join(tags)
            line = '  * Tags: {}'.format(tags_list)
            infos_lines.append(line)

        if infos_lines:
            print('\n\nInfos:\n')
            for line in infos_lines:
                print(line)

    @staticmethod
    def _print_info_episode(episode):
        emission = episode.get_emission()
        qualities = episode.get_available_qualities()

        print(emission.get_title())
        print('{}  [{}]'.format(episode.get_title(), episode.get_sae()))

        if episode.get_description() is not None:
            print('')
            description = textwrap.wrap(episode.get_description(), 80)
            for line in description:
                print(line)

        infos_lines = []
        air_date = episode.get_air_date()
        if air_date is not None:
            line = '  * Air date: {}'.format(air_date)
            infos_lines.append(line)
        infos_lines.append('  * Available qualities:')
        for quality in qualities:
            bitrate = quality.bitrate // 1000

            xres = quality.xres or '?'
            yres = quality.yres or '?'

            resolution = '{}x{}'.format(xres, yres)
            line = '    * {} kbps ({})'.format(bitrate, resolution)
            infos_lines.append(line)

        print('\n\nInfos:\n')
        for line in infos_lines:
            print(line)

    @staticmethod
    def _get_average_bitrate(qualities):
        mi = qualities[0].bitrate
        ma = qualities[-1].bitrate
        avg = (ma + mi) // 2
        closest = min(qualities, key=lambda q: abs(q.bitrate - avg))

        return closest.bitrate

    def _print_cur_pb(self, done_segments, done_bytes, force):
        if self._verbose:
            return

        cur_time = time.time()

        # ensure 100% is printed
        if done_segments == self._cur_segments_count:
            force = True

        if not force and (cur_time - self._last_pb_time < .2):
            return

        self._last_pb_time = cur_time
        bar = self._cur_pb.get_bar(done_segments, done_bytes)

        sys.stdout.write('\r{}'.format(bar))
        sys.stdout.flush()

    def _on_dl_start(self, total_segments):
        self._cur_segments_count = total_segments
        self._cur_pb = ProgressBar(self._seg_handler.filename, total_segments)
        self._last_pb_time = time.time()
        if self._quiet:
            print("Downloading {} ... ".format(self._seg_handler.filename), end="", flush=True)
        else:
            self._print_cur_pb(0, 0, True)

    def _on_dl_progress_update(self, num_completed_segments,
                               num_bytes_completed_segments,
                               num_bytes_partial_segment):
        if self._stop:
            return
        if self._quiet:
            return

        total_bytes = num_bytes_completed_segments + num_bytes_partial_segment
        self._print_cur_pb(num_completed_segments, total_bytes, False)

    @staticmethod
    def _get_fetch_filename_for_episode(episode, quality_level):
        emission_title = episode.get_emission().Title
        episode_title = episode.Title

        if episode.SeasonAndEpisode is not None:
            sae = episode.SeasonAndEpisode
            episode_title = '{} {}'.format(sae, episode_title)

        episode_title = '{}.{}'.format(episode_title, quality_level)
        filename = '{}.{}.ts'.format(emission_title, episode_title)

        # remove illegal characters from filename
        regex = r'[^ \'a-zA-Z0-9áàâäéèêëíìîïóòôöúùûüÁÀÂÄÉÈÊËÍÌÎÏÓÒÔÖÚÙÛÜçÇ()._-]'
        filename = re.sub(regex, '', filename)
        filename = re.sub(r'\s', '.', filename)

        return filename

    def _fetch_episode(self, episode, output_dir, bitrate, quality, overwrite):
        # Get available bitrates for episode
        qualities = episode.get_available_qualities()

        quality_level = "qAVG"
        # Choose bitrate
        if bitrate is None:
            if quality == App.QUALITY_MIN:
                bitrate = qualities[0].bitrate
                quality_level = "qMIN"
            elif quality == App.QUALITY_MAX:
                bitrate = qualities[-1].bitrate
                quality_level = "qMAX"
            elif quality == App.QUALITY_AVG:
                bitrate = App._get_average_bitrate(qualities)

        filename = App._get_fetch_filename_for_episode(episode, quality_level)

        # Create segment handler
        self._seg_handler = toutv.dl.FilesystemSegmentHandler(
            episode=episode, bitrate=bitrate, output_dir=output_dir, filename=filename,
            overwrite=overwrite)

        seg_provider = toutv.dl.ToutvApiSegmentProvider(
            episode=episode, bitrate=bitrate)

        # Create downloader
        self._dl = toutv.dl.Downloader(
            seg_provider=seg_provider,
            seg_handler=self._seg_handler,
            on_progress_update=self._on_dl_progress_update,
            on_dl_start=self._on_dl_start)

        # Start download
        self._dl.download()

        # Finished
        self._dl = None
        if self._quiet:
            print("Done.")

    def _fetch_emission_episodes(self, emission, output_dir, bitrate, quality, overwrite):
        episodes = self._toutv_client.get_emission_episodes(emission, True)

        if not episodes:
            title = emission.get_title()
            print('No episodes available for emission "{}"'.format(title))
            return

        for episode in App._sort_episodes(episodes):
            title = episode.get_title()

            try:
                if self._stop:
                    raise toutv.dl.CancelledByUserError()
                if episode.PID is None:
                    episode = self._toutv_client.get_episode_by_name(emission, str(episode.Id))
                self._fetch_episode(episode, output_dir, bitrate, quality, overwrite)
                sys.stdout.write('\n')
                sys.stdout.flush()
            except toutv.exceptions.RequestTimeoutError:
                tmpl = 'Error: cannot fetch "{}": request timeout'
                print(tmpl.format(title), file=sys.stderr)
            except toutv.exceptions.UnexpectedHttpStatusCodeError:
                tmpl = 'Error: cannot fetch "{}": unexpected HTTP status code'
                print(tmpl.format(title), file=sys.stderr)
            except toutv.exceptions.NetworkError as e:
                tmpl = 'Error: cannot fetch "{}": NetworkError - {}'
                print(tmpl.format(title, e), file=sys.stderr)
            except toutv.dl.FileExistsError as e:
                tmpl = 'Error: cannot fetch "{}": destination file {} already exists'
                print(tmpl.format(title, e.path), file=sys.stderr)
            except toutv.dl.CancelledByUserError as e:
                raise e
            except toutv.dl.DownloadError as e:
                tmpl = 'Error: cannot fetch "{}": DownloadError - {}'
                print(tmpl.format(title, e), file=sys.stderr)
            except Exception as e:
                tmpl = 'Error: cannot fetch "{}": {}'
                print(tmpl.format(title, e), file=sys.stderr)
                if self._verbose:
                    traceback.print_exc()

    @staticmethod
    def _sort_episodes(episodes):
        def episode_sort_func(episode):
            return distutils.version.LooseVersion(episode.get_sae())

        return sorted(episodes, key=episode_sort_func)


def _register_sigint():
    if platform.system() == 'Linux':
        def handler(signal, frame):
            print('Cancelled by user', file=sys.stderr)
            sys.exit(1)

        import signal
        signal.signal(signal.SIGINT, handler)


def run():
    app = App(sys.argv[1:])
    _register_sigint()

    return app.run()


if __name__ == '__main__':
    run()



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import shelve
from datetime import datetime
from datetime import timedelta


class Cache:

    def __init__(self):
        pass

    def get_emissions(self):
        pass

    def get_emission_episodes(self, emission_id):
        pass

    def get_page_repertoire(self):
        pass

    def set_emissions(self, emissions):
        pass

    def set_emission_episodes(self, emission_id, episodes):
        pass

    def set_page_repertoire(self, page_repertoire):
        pass

    def invalidate(self):
        pass


class EmptyCache(Cache):

    def get_emissions(self):
        return None

    def get_emission_episodes(self, emission_id):
        return None

    def get_page_repertoire(self):
        return None


class ShelveCache(Cache):

    _cache_version = 4

    def __init__(self, shelve_filename):
        self._logger = logging.getLogger(self.__class__.__name__)

        try:
            self._logger.debug('Trying to open shelve at {}'.format(shelve_filename))
            self.shelve = shelve.open(shelve_filename)

            if ('cache_version' not in self.shelve or
                    self.shelve['cache_version'] != self._cache_version):
                self._logger.debug('Incompatible cache version, invalidating.')
                self.invalidate()

        except Exception as e:
            self.shelve = None
            raise e

    def __del__(self):
        if self.shelve is not None:
            self.shelve.close()

    def _has_key(self, key):
        if key in self.shelve:
            expire, value = self.shelve[key]
            if datetime.now() < expire:
                return True

        return False

    def _get(self, key):
        if not self._has_key(key):
            return None

        expire, value = self.shelve[key]

        return value

    def _set(self, key, value, expire=timedelta(hours=2)):
        self.shelve[key] = (datetime.now() + expire, value)

    def _del(self, key):
        if key in self.shelve:
            del self.shelve[key]

    def get_emissions(self):
        return self._get('emissions')

    def get_emission_episodes(self, emission):
        emid = emission.Id
        emission_episodes = self._get('emission_episodes')
        if emission_episodes is None:
            return None
        if emid not in emission_episodes:
            return None

        return emission_episodes[emid]

    def get_page_repertoire(self):
        return self._get('page_repertoire')

    def set_emissions(self, emissions):
        self._set('emissions', emissions)

    def set_emission_episodes(self, emission, episodes):
        emid = emission.Id
        emission_episodes = self._get('emission_episodes')
        if emission_episodes is None:
            emission_episodes = {}
        emission_episodes[emid] = episodes
        self._set('emission_episodes', emission_episodes)

    def set_page_repertoire(self, page_repertoire):
        self._set('page_repertoire', page_repertoire)

    def invalidate(self):
        self._del('emissions')
        self._del('emission_episodes')
        self._del('page_repertoire')
        self.shelve['cache_version'] = self._cache_version
        self.shelve.sync()



import unittest
from toutv import dl


class DummySegmentProvider(dl.SegmentProvider):

    def __init__(self):
        self._segments = [
            b'abcd',
            b'efgh',
            b'ijkl',
            b'mnop',
        ]

    def initialize(self):
        pass

    def download_segment(self, segindex, progress):
        # Simulate that we have a partial progress update where we have
        # downloaded 2 of the 4 bytes.
        progress(2)

        return self._segments[segindex]

    def num_segments(self):
        return len(self._segments)

    def finalize(self):
        pass


class DummySegmentHandler(dl.SegmentHandler):

    def __init__(self):
        self._segments = []

    def initialize(self):
        pass

    def finalize(self, num_segments):
        pass

    def segment_size(self, segindex):
        pass

    def has_segment(self, segindex):
        return False

    def on_segment(self, segindex, segment):
        assert segindex == len(self._segments)

        self._segments.append(segment)


class DlTest(unittest.TestCase):

    def test_simple_download(self):
        seg_provider = DummySegmentProvider()
        seg_handler = DummySegmentHandler()
        downloader = dl.Downloader(seg_provider, seg_handler)
        downloader.download()
        assert seg_provider._segments == seg_handler._segments

    def test_on_progress_update(self):

        seg_provider = DummySegmentProvider()
        seg_handler = DummySegmentHandler()

        class Progress:

            _expected_values = [
                # completed segments, bytes of completed segments, bytes of current partial segment
                (0, 0, 0),
                (0, 0, 2),
                (1, 4, 0),
                (1, 4, 2),
                (2, 8, 0),
                (2, 8, 2),
                (3, 12, 0),
                (3, 12, 2),
                (4, 16, 0),
            ]

            def __init__(self):
                self._num_calls = 0

            def progress(self,
                         num_completed_segments,
                         num_bytes_completed_segments,
                         num_bytes_partial_segment):
                exp = self._expected_values[self._num_calls]
                assert exp == (num_completed_segments, num_bytes_completed_segments, num_bytes_partial_segment)
                self._num_calls += 1

        p = Progress()

        downloader = dl.Downloader(seg_provider, seg_handler, on_progress_update=p.progress)
        downloader.download()



# Copyright (c) 2015, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# All rights reserved.
#
# Gladly lifted from my friend Simon Carpentier <simon.carpentier@spacebar.ca>
#   https://github.com/scarpentier/toutv-extra
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn OR Philippe Proulx
# BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
import requests
import toutv.config
import toutv.exceptions


class Auth:

    def __init__(self, token=None):
        self._token = token
        self._claims = None

    def get_claims(self, token):
        if not self._claims:
            headers = {
                "Authorization": "Bearer " + token,
                "User-Agent": toutv.config.USER_AGENT,
                "Accept": "application/json",
                "Host": "services.radio-canada.ca",
            }

            r = requests.get(toutv.config.TOUTV_AUTH_CLAIMS_URL.format(token), headers=headers)

            if r.status_code != 200:
                raise toutv.exceptions.UnexpectedHttpStatusCodeError(toutv.config.TOUTV_AUTH_CLAIMS_URL.format(token), r.status_code)

            self._claims = r.json()["claims"]
        return self._claims

    def get_token(self):
        return self._token

    def login(self, username, password):
        session = self._get_session()

        payload = {
            "action": "login",
            "sessionID": session['sessionID'],
            "sessionData": session['sessionData'],
            "authzRequestUri": session['authzRequestUri'],
            "lang": session['lang'],
            "login-email": username,
            "login-password": password,
            "form-submit-btn": "Ouvrir une session"
        }

        headers = {
            "User-Agent": toutv.config.USER_AGENT,
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Origin": "https://services.radio-canada.ca",
            "X-Requested-With": "tv.tou.android",
            "Referer": toutv.config.TOUTV_AUTH_SESSION_URL,
            "Content-type": "application/x-www-form-urlencoded",
        }

        r = requests.post(toutv.config.TOUTV_AUTH_LOGIN_URL, headers=headers, data=payload, allow_redirects=False)

        if r.status_code != 200:
            raise toutv.exceptions.UnexpectedHttpStatusCodeError(toutv.config.TOUTV_AUTH_TOKEN_URL, r.status_code)

        headers['Referer'] = toutv.config.TOUTV_AUTH_LOGIN_URL

        payload = {
            "action": "grant",
            "sessionID": session['sessionID'],
            "sessionData": re.search("name=\"sessionData\" .*value=\"([^\"]*)\"", r.text).group(1),
            "lang": session['lang']
        }

        r = requests.post(toutv.config.TOUTV_AUTH_CONSENT_URL, headers=headers, data=payload, allow_redirects=False)

        if r.status_code != 302:
            raise toutv.exceptions.UnexpectedHttpStatusCodeError(toutv.config.TOUTV_AUTH_TOKEN_URL, r.status_code)

        self._token = re.search("access_token=([^&]*)", r.headers["Location"]).group(1)

    def _get_session(self):
        headers = {
            "User-Agent": toutv.config.USER_AGENT,
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Origin": "https://services.radio-canada.ca",
            "X-Requested-With": "tv.tou.android"
        }

        r = requests.get(toutv.config.TOUTV_AUTH_SESSION_URL, headers=headers)

        if r.status_code != 200:
            raise toutv.exceptions.UnexpectedHttpStatusCodeError(toutv.config.TOUTV_AUTH_SESSION_URL, r.status_code)

        session = {
            'sessionID': re.search("name=\"sessionID\" .*value=\"([^\"]*)\"", r.text).group(1),
            'sessionData': re.search("name=\"sessionData\" .*value=\"([^\"]*)\"", r.text).group(1),
            'authzRequestUri': re.search("name=\"authzRequestUri\" .*value=\"([^\"]*)\"", r.text).group(1),
            'lang': re.search("name=\"lang\" .*value=\"([^\"]*)\"", r.text).group(1)
        }

        return session
